from __future__ import annotations

__docformat__ = "numpy"

from collections.abc import Callable
import logging
import os
import os.path as op
import random
import re
import shutil
import sys
from types import ModuleType
from typing import TYPE_CHECKING, Any, List, Optional, cast

import filelock
from nipype import Node
from nipype.interfaces.base import TraitListObject

from .bids import (
    BIDS_VERSION,
    BIDSError,
    add_participant_record,
    populate_bids_templates,
    populate_intended_for,
    sanitize_label,
    save_scans_key,
    tuneup_bids_json_files,
)
from .dicoms import (
    compress_dicoms,
    embed_metadata_from_dicoms,
    group_dicoms_into_seqinfos,
)
from .due import Doi, due
from .utils import (
    SeqInfo,
    TempDirs,
    assure_no_file_exists,
    clear_temp_dicoms,
    file_md5sum,
    load_json,
    read_config,
    safe_copyfile,
    safe_movefile,
    save_json,
    set_readonly,
    treat_infofile,
    write_config,
)

if TYPE_CHECKING:
    if sys.version_info >= (3, 8):
        from typing import TypedDict
    else:
        from typing_extensions import TypedDict

    class PopulateIntendedForOpts(TypedDict, total=False):
        matching_parameters: str | list[str]
        criterion: str


LOCKFILE = "heudiconv.lock"
DW_IMAGE_IN_FMAP_FOLDER_WARNING = (
    "Diffusion-weighted image saved in non dwi folder ({folder})"
)
lgr = logging.getLogger(__name__)


def conversion_info(
    subject: str,
    outdir: str,
    info: dict[tuple[str, tuple[str, ...], None], list],
    filegroup: dict[str, list[str]],
    ses: Optional[str],
) -> list[tuple[str, tuple[str, ...], list[str]]]:
    convert_info: list[tuple[str, tuple[str, ...], list[str]]] = []
    for key, items in info.items():
        if not items:
            continue
        template, outtype = key[0], key[1]
        # So no annotation_classes of any kind!  so if not used -- what was the
        # intention???? XXX
        outpath = outdir
        for idx, itemgroup in enumerate(items):
            if not isinstance(itemgroup, list):
                itemgroup = [itemgroup]
            for subindex, item in enumerate(itemgroup):
                parameters = {}
                if isinstance(item, dict):
                    parameters = {k: v for k, v in item.items()}
                    item = parameters["item"]
                    del parameters["item"]
                # some helper meta-varaibles
                parameters.update(
                    dict(
                        item=idx + 1,
                        subject=subject,
                        seqitem=item,
                        subindex=subindex + 1,
                        session="ses-" + str(ses),
                        bids_subject_session_prefix="sub-%s" % subject
                        + (("_ses-%s" % ses) if ses else ""),
                        bids_subject_session_dir="sub-%s" % subject
                        + (("/ses-%s" % ses) if ses else ""),
                        # referring_physician_name
                        # study_description
                    )
                )
                try:
                    files = filegroup[item]
                except KeyError:
                    files = filegroup[str(item)]
                outprefix = template.format(**parameters)
                convert_info.append((op.join(outpath, outprefix), outtype, files))
    return convert_info


def prep_conversion(
    sid: Optional[str],
    dicoms: Optional[list[str]],
    outdir: str,
    heuristic: ModuleType,
    converter: str,
    anon_sid: Optional[str],
    anon_outdir: Optional[str],
    with_prov: bool,
    ses: Optional[str],
    bids_options: Optional[str],
    seqinfo: Optional[dict[SeqInfo, list[str]]],
    min_meta: bool,
    overwrite: bool,
    dcmconfig: Optional[str],
    grouping: str,
) -> None:
    if dicoms:
        lgr.info("Processing %d dicoms", len(dicoms))
    elif seqinfo:
        lgr.info("Processing %d pre-sorted seqinfo entries", len(seqinfo))
    else:
        raise ValueError("neither dicoms nor seqinfo dict was provided")

    if bids_options is not None:
        if not sid:
            raise ValueError(
                "BIDS requires alphanumeric subject ID. Got an empty value"
            )
        sid = sanitize_label(sid)
        if ses:
            ses = sanitize_label(ses)

    if not anon_sid:
        if sid is None:
            raise ValueError("Neither 'sid' nor 'anon_sid' is true")
        anon_sid = sid
    if not anon_outdir:
        anon_outdir = outdir

    # Generate heudiconv info folder
    idir = op.join(outdir, ".heudiconv", anon_sid)
    if bids_options is not None and ses:
        idir = op.join(idir, "ses-%s" % str(ses))
    if anon_outdir == outdir:
        idir = op.join(idir, "info")
    if not op.exists(idir):
        os.makedirs(idir)

    ses_suffix = "_ses-%s" % ses if ses is not None else ""
    info_file = op.join(idir, "%s%s.auto.txt" % (sid, ses_suffix))
    edit_file = op.join(idir, "%s%s.edit.txt" % (sid, ses_suffix))
    filegroup_file = op.join(idir, "filegroup%s.json" % ses_suffix)

    # if conversion table(s) do not exist -- we need to prepare them
    # (the *prepare* stage in https://github.com/nipy/heudiconv/issues/134)
    # if overwrite - recalculate this anyways
    reuse_conversion_table = op.exists(edit_file)
    # We also might need to redo it if changes in the heuristic file
    # detected
    # ref: https://github.com/nipy/heudiconv/issues/84#issuecomment-330048609
    # for more automagical wishes
    target_heuristic_filename = op.join(idir, "heuristic.py")
    # facilitates change - TODO: remove in 1.0
    old_heuristic_filename = op.join(idir, op.basename(heuristic.filename))
    if op.exists(old_heuristic_filename):
        assure_no_file_exists(target_heuristic_filename)
        safe_copyfile(old_heuristic_filename, target_heuristic_filename)
        assure_no_file_exists(old_heuristic_filename)
    # TODO:
    #  1. add a test
    #  2. possibly extract into a dedicated function for easier logic flow here
    #     and a dedicated unittest
    if op.exists(target_heuristic_filename) and file_md5sum(
        target_heuristic_filename
    ) != file_md5sum(heuristic.filename):
        # remake conversion table
        reuse_conversion_table = False
        lgr.info(
            "Will not reuse existing conversion table files because heuristic "
            "has changed"
        )

    info: dict[tuple[str, tuple[str, ...], None], list]
    if reuse_conversion_table:
        lgr.info("Reloading existing filegroup.json " "because %s exists", edit_file)
        info = read_config(edit_file)
        filegroup = load_json(filegroup_file)
        # XXX Yarik finally understood why basedir was dragged along!
        # So we could reuse the same PATHs definitions possibly consistent
        # across re-runs... BUT that wouldn't work anyways if e.g.
        # DICOMs dumped with SOP UUIDs thus differing across runs etc
        # So either it would need to be brought back or reconsidered altogether
        # (since no sample data to test on etc)
    else:
        assure_no_file_exists(target_heuristic_filename)
        safe_copyfile(heuristic.filename, target_heuristic_filename)
        if dicoms:
            seqinfo = group_dicoms_into_seqinfos(
                dicoms,
                grouping,
                file_filter=getattr(heuristic, "filter_files", None),
                dcmfilter=getattr(heuristic, "filter_dicom", None),
                flatten=True,
                custom_grouping=getattr(heuristic, "grouping", None),
                # callable which will be provided dcminfo and returned
                # structure extend seqinfo
                custom_seqinfo=getattr(heuristic, "custom_seqinfo", None),
            )
        elif seqinfo is None:
            raise ValueError("Neither 'dicoms' nor 'seqinfo' is given")

        seqinfo_list = list(seqinfo.keys())
        filegroup = {si.series_id: x for si, x in seqinfo.items()}
        dicominfo_file = op.join(idir, "dicominfo%s.tsv" % ses_suffix)
        # allow to overwrite even if was present under git-annex already
        assure_no_file_exists(dicominfo_file)
        with open(dicominfo_file, "wt") as fp:
            fp.write("\t".join(SeqInfo._fields) + "\n")
            for seq in seqinfo_list:
                fp.write("\t".join([str(val) for val in seq]) + "\n")
        lgr.debug("Calling out to %s.infodict", heuristic)
        info = heuristic.infotodict(seqinfo_list)
        lgr.debug("Writing to {}, {}, {}".format(info_file, edit_file, filegroup_file))
        assure_no_file_exists(info_file)
        write_config(info_file, info)
        assure_no_file_exists(edit_file)
        write_config(edit_file, info)
        save_json(filegroup_file, filegroup)

    if bids_options is not None:
        # the other portion of the path would mimic BIDS layout
        # so we don't need to worry here about sub, ses at all
        tdir = anon_outdir
    else:
        tdir = op.join(anon_outdir, anon_sid)

    if converter.lower() != "none":
        lgr.info("Doing conversion using %s", converter)
        cinfo = conversion_info(anon_sid, tdir, info, filegroup, ses)
        convert(
            cinfo,
            converter=converter,
            scaninfo_suffix=getattr(heuristic, "scaninfo_suffix", ".json"),
            custom_callable=getattr(heuristic, "custom_callable", None),
            populate_intended_for_opts=getattr(
                heuristic, "POPULATE_INTENDED_FOR_OPTS", None
            ),
            with_prov=with_prov,
            bids_options=bids_options,
            outdir=tdir,
            min_meta=min_meta,
            overwrite=overwrite,
            dcmconfig=dcmconfig,
        )

    for item_dicoms in filegroup.values():
        clear_temp_dicoms(item_dicoms)

    if bids_options is not None and "notop" not in bids_options:
        lockfile = op.join(anon_outdir, LOCKFILE)
        if op.exists(lockfile):
            lgr.warning(
                "Existing lockfile found in {0} - waiting for the "
                "lock to be released. To set a timeout limit, set "
                "the HEUDICONV_FILELOCK_TIMEOUT environmental variable "
                "to a value in seconds. If this process hangs, it may "
                "require a manual deletion of the {0}.".format(lockfile)
            )
        timeout = float(os.getenv("HEUDICONV_LOCKFILE_TIMEOUT", -1))
        with filelock.SoftFileLock(lockfile, timeout=timeout):
            if seqinfo:
                keys = list(seqinfo)
                add_participant_record(
                    anon_outdir, anon_sid, keys[0].patient_age, keys[0].patient_sex
                )
            populate_bids_templates(
                anon_outdir, getattr(heuristic, "DEFAULT_FIELDS", {})
            )


def update_complex_name(metadata: dict[str, Any], filename: str) -> str:
    """
    Insert `_part-<mag|phase>` entity into filename if data are from a
    sequence with magnitude/phase part.

    Parameters
    ----------
    metadata : dict
        Scan metadata dictionary from BIDS sidecar file.
    filename : str
        Incoming filename

    Returns
    -------
    filename : str
        Updated filename with part entity added in appropriate position.
    """
    # Some scans separate magnitude/phase differently
    # A small note: _phase is deprecated, but this may add part-mag to
    # magnitude data while leaving phase data with a separate suffix,
    # depending on how one sets up their heuristic.
    unsupported_types = [
        "_phase",
        "_magnitude",
        "_magnitude1",
        "_magnitude2",
        "_phasediff",
        "_phase1",
        "_phase2",
    ]
    if any(ut in filename for ut in unsupported_types):
        return filename

    # Check to see if it is magnitude or phase part:
    img_type = cast(List[str], metadata.get("ImageType", []))
    if "M" in img_type:
        mag_or_phase = "mag"
    elif "P" in img_type:
        mag_or_phase = "phase"
    else:
        raise RuntimeError("Data type could not be inferred from the metadata.")

    # Determine scan suffix
    filetype = "_" + filename.split("_")[-1]

    # Insert part label
    if not ("_part-%s" % mag_or_phase) in filename:
        # If "_part-" is specified, prepend the 'mag_or_phase' value.
        if "_part-" in filename:
            raise BIDSError(
                "Part label for images will be automatically set, "
                "remove from heuristic"
            )

        # Insert it **before** the following string(s), whichever appears first.
        # https://bids-specification.readthedocs.io/en/stable/99-appendices/09-entities.html
        entities_after_part = [
            "_proc",
            "_hemi",
            "_space",
            "_split",
            "_recording",
            "_chunk",
            "_res",
            "_den",
            "_label",
            "_desc",
            filetype,
        ]
        for label in entities_after_part:
            if (label == filetype) or (label in filename):
                filename = filename.replace(label, "_part-%s%s" % (mag_or_phase, label))
                break

    return filename


def update_multiecho_name(
    metadata: dict[str, Any], filename: str, echo_times: list[float]
) -> str:
    """
    Insert `_echo-<num>` entity into filename if data are from a multi-echo
    sequence.

    Parameters
    ----------
    metadata : dict
        Scan metadata dictionary from BIDS sidecar file.
    filename : str
        Incoming filename
    echo_times : list
        List of all echo times from scan. Used to determine the echo *number*
        (i.e., index) if field is missing from metadata.

    Returns
    -------
    filename : str
        Updated filename with echo entity added, if appropriate.
    """
    # Field maps separate echoes differently, so do not attempt to update any filenames with these
    # suffixes
    unsupported_types = [
        "_magnitude",
        "_magnitude1",
        "_magnitude2",
        "_phasediff",
        "_phase1",
        "_phase2",
        "_fieldmap",
    ]
    if any(ut in filename for ut in unsupported_types):
        return filename

    if not isinstance(echo_times, list):
        raise TypeError(
            f'Argument "echo_times" must be a list, not a {type(echo_times)}'
        )

    # Get the EchoNumber from json file info.  If not present, use EchoTime.
    if "EchoNumber" in metadata.keys():
        echo_number = metadata["EchoNumber"]
        assert isinstance(echo_number, int)
    elif "EchoTime" in metadata.keys():
        echo_number = echo_times.index(metadata["EchoTime"]) + 1
    else:
        raise KeyError(
            'Either "EchoNumber" or "EchoTime" must be in metadata keys. '
            f"Keys detected: {metadata.keys()}"
        )

    # Determine scan suffix
    filetype = "_" + filename.split("_")[-1]

    # Insert it **before** the following string(s), whichever appears first.
    # https://bids-specification.readthedocs.io/en/stable/99-appendices/09-entities.html
    entities_after_echo = [
        "_flip",
        "_inv",
        "_mt",
        "_part",
        "_proc",
        "_hemi",
        "_space",
        "_split",
        "_recording",
        "_chunk",
        "_res",
        "_den",
        "_label",
        "_desc",
        filetype,
    ]
    for label in entities_after_echo:
        if (label == filetype) or (label in filename):
            filename = filename.replace(label, "_echo-%s%s" % (echo_number, label))
            break

    return filename


def update_uncombined_name(
    metadata: dict[str, Any], filename: str, channel_names: list[str]
) -> str:
    """
    Insert `_ch-<num>` entity into filename if data are from a sequence
    with "save uncombined".

    Parameters
    ----------
    metadata : dict
        Scan metadata dictionary from BIDS sidecar file.
    filename : str
        Incoming filename
    channel_names : list
        List of all channel names from scan. Used to determine the channel
        *number* (i.e., index) if field is missing from metadata.

    Returns
    -------
    filename : str
        Updated filename with ch entity added, if appropriate.
    """
    # In case any scan types separate channels differently
    unsupported_types: list[str] = []
    if any(ut in filename for ut in unsupported_types):
        return filename

    if not isinstance(channel_names, list):
        raise TypeError(
            f'Argument "channel_names" must be a list, not a {type(channel_names)}'
        )

    # Determine the channel number
    coil_string = metadata["CoilString"]
    assert isinstance(coil_string, str)
    channel_number = "".join(c for c in coil_string if c.isdigit())
    if not channel_number:
        channel_number = str(channel_names.index(coil_string) + 1)
    channel_number = channel_number.zfill(2)

    # Determine scan suffix
    filetype = "_" + filename.split("_")[-1]

    # Insert it **before** the following string(s), whichever appears first.
    # Choosing to put channel near the end since it's not in the specification yet.
    # See https://bids-specification.readthedocs.io/en/stable/99-appendices/09-entities.html
    entities_after_ch = [
        "_proc",
        "_hemi",
        "_space",
        "_split",
        "_recording",
        "_chunk",
        "_res",
        "_den",
        "_label",
        "_desc",
        filetype,
    ]
    for label in entities_after_ch:
        if (label == filetype) or (label in filename):
            filename = filename.replace(label, "_ch-%s%s" % (channel_number, label))
            break
    return filename


def convert(
    items: list[tuple[str, tuple[str, ...], list[str]]],
    converter: str,
    scaninfo_suffix: str,
    custom_callable: Optional[Callable[[str, tuple[str, ...], list[str]], Any]],
    with_prov: bool,
    bids_options: Optional[str],
    outdir: str,
    min_meta: bool,
    overwrite: bool,
    symlink: bool = True,
    prov_file: Optional[str] = None,
    dcmconfig: Optional[str] = None,
    populate_intended_for_opts: Optional[PopulateIntendedForOpts] = None,
) -> None:
    """Perform actual conversion (calls to converter etc) given info from
    heuristic's `infotodict`
    """
    prov_files: list[str] = []
    tempdirs = TempDirs()

    if bids_options is not None:
        due.cite(
            # doi matches the BIDS_VERSION
            Doi("10.5281/zenodo.4085321"),
            description="Brain Imaging Data Structure (BIDS) Specification",
            path="bids",
            version=BIDS_VERSION,
            tags=["implementation"],
        )
        due.cite(
            Doi("10.1038/sdata.2016.44"),
            description="Brain Imaging Data Structure (BIDS), Original paper",
            path="bids",
            tags=["documentation"],
        )

    for item in items:
        prefix, outtypes, item_dicoms = item
        prefix_dirname = op.dirname(prefix)
        outname_bids = prefix + ".json"
        bids_outfiles = []
        # set empty outname and scaninfo in case we only want dicoms
        outname = ""
        scaninfo = ""
        lgr.info(
            "Converting %s (%d DICOMs) -> %s . Converter: %s . Output types: %s",
            prefix,
            len(item_dicoms),
            prefix_dirname,
            converter,
            outtypes,
        )
        # We want to create this dir only if we are converting it to nifti,
        # or if we're using BIDS
        dicom_only = outtypes == ("dicom",)
        if not (dicom_only and (bids_options is not None)) and not op.exists(
            prefix_dirname
        ):
            os.makedirs(prefix_dirname)

        for outtype in outtypes:
            lgr.debug(
                "Processing %d dicoms for output type %s. Overwrite=%s",
                len(item_dicoms),
                outtype,
                overwrite,
            )
            lgr.debug("Includes the following dicoms: %s", item_dicoms)

            if outtype == "dicom":
                convert_dicom(
                    item_dicoms,
                    bids_options,
                    prefix,
                    outdir,
                    tempdirs,
                    symlink,
                    overwrite,
                )
            elif outtype in ["nii", "nii.gz"]:
                assert converter == "dcm2niix", f"Invalid converter {converter}"
                due.cite(
                    Doi("10.1016/j.jneumeth.2016.03.001"),
                    path="dcm2niix",
                    description="DICOM to NIfTI + .json sidecar conversion utility",
                    tags=["implementation"],
                )
                outname, scaninfo = (prefix + "." + outtype, prefix + scaninfo_suffix)

                if not op.exists(outname) or overwrite:
                    tmpdir = tempdirs("dcm2niix")

                    # run conversion through nipype
                    res, prov_file = nipype_convert(
                        item_dicoms, prefix, with_prov, bids_options, tmpdir, dcmconfig
                    )

                    bids_outfiles = save_converted_files(
                        res,
                        item_dicoms,
                        bids_options,
                        outtype,
                        prefix,
                        outname_bids,
                        overwrite=overwrite,
                    )

                    # save acquisition time information if it's BIDS
                    # at this point we still have acquisition date
                    if bids_options is not None:
                        save_scans_key(item, bids_outfiles)
                    # Fix up and unify BIDS files
                    tuneup_bids_json_files(bids_outfiles)

                    if prov_file:
                        prov_files.append(prov_file)

                    tempdirs.rmtree(tmpdir)
                else:
                    raise RuntimeError(
                        "was asked to convert into %s but destination already exists"
                        % (outname)
                    )

        # add the taskname field to the json file(s):
        add_taskname_to_infofile(bids_outfiles)

        if len(bids_outfiles) > 1:
            lgr.warning(
                "For now not embedding BIDS and info generated "
                ".nii.gz itself since sequence produced "
                "multiple files"
            )
        elif not bids_outfiles:
            lgr.debug("No BIDS files were produced, nothing to embed to then")
        elif outname and not min_meta:
            embed_metadata_from_dicoms(
                bids_options,
                item_dicoms,
                outname,
                outname_bids,
                prov_file,
                scaninfo,
                tempdirs,
                with_prov,
            )
        if scaninfo and op.exists(scaninfo):
            lgr.info("Post-treating %s file", scaninfo)
            treat_infofile(scaninfo)

        # this may not always be the case: ex. fieldmap1, fieldmap2
        # will address after refactor
        if outname and op.exists(outname):
            set_readonly(outname)

        if custom_callable is not None:
            custom_callable(*item)

    # Populate "IntendedFor" for fmap files if requested in heuristic
    if populate_intended_for_opts is not None:
        # Because fmap files can only be used to correct for distortions in images
        # collected within the same scanning session, find unique subject/session
        # combinations from the outname in each item:
        outnames = [item[0] for item in items]
        # - grab "sub-<sID>[/ses-<ses>]", and keep only unique ones:
        sessions: set[str] = set()
        for oname in outnames:
            m = re.search(
                "sub-(?P<subj>[a-zA-Z0-9]*)([{0}_]ses-(?P<ses>[a-zA-Z0-9]*))?".format(
                    op.sep
                ),
                oname,
            )
            if m:
                sessions.add(m.group(0))
            else:
                # "sub-<sID>[/ses-<ses>]" is not present, so this is not BIDS
                # compliant and it doesn't make sense to add "IntendedFor":
                sessions.clear()
                break

        for ses in sessions:
            session_path = op.join(outdir, ses)
            populate_intended_for(session_path, **populate_intended_for_opts)


def convert_dicom(
    item_dicoms: list[str],
    bids_options: Optional[str],
    prefix: str,
    outdir: str,
    tempdirs: TempDirs,
    _symlink: bool,
    overwrite: bool,
) -> None:
    """Save DICOMs as output (default is by symbolic link)

    Parameters
    ----------
    item_dicoms : list of filenames
        DICOMs to save
    bids_options : str or None
        If not None then save to BIDS format. String may be empty
        or contain bids specific options
    prefix : string
        Conversion outname
    outdir : string
        Output directory
    tempdirs : TempDirs instance
        Object to handle temporary directories created
        TODO: remove
    symlink : bool
        Create softlink to DICOMs - if False, create hardlink instead.
    overwrite : bool
        If True, allows overwriting of previous conversion
    """
    if bids_options is not None:
        # mimic the same hierarchy location as the prefix
        # although it could all have been done probably
        # within heuristic really
        sourcedir = op.join(outdir, "sourcedata")
        sourcedir_ = op.join(sourcedir, op.dirname(op.relpath(prefix, outdir)))
        if not op.exists(sourcedir_):
            os.makedirs(sourcedir_)

        compress_dicoms(
            item_dicoms, op.join(sourcedir_, op.basename(prefix)), tempdirs, overwrite
        )
    else:
        dicomdir = prefix + "_dicom"
        if op.exists(dicomdir):
            lgr.info(
                "Found existing DICOM directory {}, " "removing...".format(dicomdir)
            )
            shutil.rmtree(dicomdir)
        os.mkdir(dicomdir)
        for filename in item_dicoms:
            outfile = op.join(dicomdir, op.basename(filename))
            if not op.islink(outfile):
                # TODO: add option to enable hardlink?
                #                if symlink:
                #                    os.symlink(filename, outfile)
                #                else:
                #                    os.link(filename, outfile)
                shutil.copyfile(filename, outfile)


def nipype_convert(
    item_dicoms: list[str],
    prefix: str,
    with_prov: bool,
    bids_options: Optional[str],
    tmpdir: str,
    dcmconfig: Optional[str] = None,
) -> tuple[Node, Optional[str]]:
    """
    Converts DICOMs grouped from heuristic using Nipype's Dcm2niix interface.

    Parameters
    ----------
    item_dicoms : list
        DICOM files to convert
    prefix : str
        Heuristic output path
    with_prov : bool
        Store provenance information
    bids_options : str or None
        If not None then output BIDS sidecar JSONs
        String may contain bids specific options
    tmpdir : str
        Conversion working directory
    dcmconfig : str, optional
        JSON file used for additional Dcm2niix configuration
    """
    import nipype

    if with_prov:
        from nipype import config

        config.enable_provenance()
    from nipype.interfaces.dcm2nii import Dcm2niix

    # <https://github.com/python/mypy/issues/9864>
    item_dicoms = list(map(op.abspath, item_dicoms))  # type: ignore[arg-type]

    fromfile = dcmconfig if dcmconfig else None
    if fromfile:
        lgr.info("Using custom config file %s", fromfile)

    convertnode = Node(Dcm2niix(from_file=fromfile), name="convert")
    convertnode.base_dir = tmpdir
    convertnode.inputs.source_names = item_dicoms
    convertnode.inputs.out_filename = op.basename(
        prefix
    ) + "_heudiconv%03d" % random.randint(0, 999)
    prefix_dir = op.dirname(prefix)
    # if provided prefix had a path in it -- pass is as output_dir instead of default curdir
    if prefix_dir:
        convertnode.inputs.output_dir = prefix_dir

    if nipype.__version__.split(".")[0] == "0":
        # deprecated since 1.0, might be needed(?) before
        convertnode.inputs.terminal_output = "allatonce"
    else:
        convertnode.terminal_output = "allatonce"
    convertnode.inputs.bids_format = bids_options is not None
    eg = convertnode.run()

    # prov information
    prov_file = prefix + "_prov.ttl" if with_prov else None
    if prov_file:
        safe_movefile(
            op.join(convertnode.base_dir, convertnode.name, "provenance.ttl"), prov_file
        )

    return eg, prov_file


def save_converted_files(
    res: Node,
    item_dicoms: list[str],
    bids_options: Optional[str],
    outtype: str,
    prefix: str,
    outname_bids: str,
    overwrite: bool,
) -> list[str]:
    """Copy converted files from tempdir to output directory.

    Will rename files if necessary.

    Parameters
    ----------
    res : Node
        Nipype conversion Node with results
    item_dicoms: list
        Filenames of converted DICOMs
    bids : list or None
        If not list save to BIDS
        List may contain bids specific options
    prefix : str

    Returns
    -------
    bids_outfiles
        Converted BIDS files

    """
    from nipype.interfaces.base import isdefined

    prefix_dirname, prefix_basename = op.split(prefix)

    bids_outfiles: list[str] = []
    res_files = res.outputs.converted_files

    if not len(res_files):
        lgr.debug("DICOMs {} were not converted".format(item_dicoms))
        return []

    if isdefined(res.outputs.bvecs) and isdefined(res.outputs.bvals):
        bvals, bvecs = res.outputs.bvals, res.outputs.bvecs
        bvals = list(bvals) if isinstance(bvals, TraitListObject) else bvals
        bvecs = list(bvecs) if isinstance(bvecs, TraitListObject) else bvecs
        if prefix_dirname.endswith("dwi"):
            outname_bvecs, outname_bvals = prefix + ".bvec", prefix + ".bval"
            safe_movefile(bvecs, outname_bvecs, overwrite)
            safe_movefile(bvals, outname_bvals, overwrite)
        else:
            if bvals_are_zero(bvals):
                to_remove = bvals + bvecs if isinstance(bvals, list) else [bvals, bvecs]
                for ftr in to_remove:
                    os.remove(ftr)
                lgr.debug("%s and %s were removed since not dwi", bvecs, bvals)
            else:
                lgr.warning(
                    DW_IMAGE_IN_FMAP_FOLDER_WARNING.format(folder=prefix_dirname)
                )
                lgr.warning(
                    ".bvec and .bval files will be generated. This is NOT BIDS compliant"
                )
                outname_bvecs, outname_bvals = prefix + ".bvec", prefix + ".bval"
                safe_movefile(bvecs, outname_bvecs, overwrite)
                safe_movefile(bvals, outname_bvals, overwrite)

    if isinstance(res_files, list):
        res_files = sorted(res_files)
        # we should provide specific handling for fmap,
        # dwi etc which might spit out multiple files

        suffixes = (
            [str(i + 1) for i in range(len(res_files))]
            if (bids_options is not None)
            else None
        )

        if not suffixes:
            lgr.warning(
                "Following series files likely have "
                "multiple (%d) volumes (orientations?) "
                "generated: %s ...",
                len(res_files),
                item_dicoms[0],
            )
            suffixes = [str(-i - 1) for i in range(len(res_files))]

        # Also copy BIDS files although they might need to
        # be merged/postprocessed later
        bids_files = (
            sorted(res.outputs.bids)
            if len(res.outputs.bids) == len(res_files)
            else [None] * len(res_files)
        )
        # preload since will be used in multiple spots
        bids_metas = [load_json(b) for b in bids_files if b]

        ###   Do we have a multi-echo series?   ###
        #   Some Siemens sequences (e.g. CMRR's MB-EPI) set the label 'TE1',
        #   'TE2', etc. in the 'ImageType' field. However, other seqs do not
        #   (e.g. MGH ME-MPRAGE). They do set a 'EchoNumber', but not for the
        #   first echo.  To compound the problem, the echoes are NOT in order,
        #   so the first NIfTI file does not correspond to echo-1, etc. So, we
        #   need to know, beforehand, whether we are dealing with a multi-echo
        #   series. To do that, the most straightforward way is to read the
        #   echo times for all bids_files and see if they are all the same or not.

        # Collect some metadata across all images
        echo_times: set[float] = set()
        channel_names: set[str] = set()
        image_types: set[str] = set()
        for metadata in bids_metas:
            if not metadata:
                continue
            try:
                echo_times.add(metadata["EchoTime"])
            except KeyError:
                pass
            try:
                channel_names.add(metadata["CoilString"])
            except KeyError:
                pass
            try:
                image_types.update(metadata["ImageType"])
            except KeyError:
                pass

        is_multiecho = (
            len(set(filter(bool, echo_times))) > 1
        )  # Check for varying echo times
        is_uncombined = (
            len(set(filter(bool, channel_names))) > 1
        )  # Check for uncombined data
        is_complex = (
            "M" in image_types and "P" in image_types
        )  # Determine if data are complex (magnitude + phase)
        echo_times_lst = sorted(echo_times)  # also converts to list
        channel_names_lst = sorted(channel_names)  # also converts to list

        ### Loop through the bids_files, set the output name and save files
        for fl, suffix, bids_file, bids_meta in zip(
            res_files, suffixes, bids_files, bids_metas
        ):
            # TODO: monitor conversion duration

            # set the prefix basename for this specific file (we'll modify it,
            # and we don't want to modify it for all the bids_files):
            this_prefix_basename = prefix_basename

            # Update name for certain criteria
            if bids_file:
                if is_multiecho:
                    this_prefix_basename = update_multiecho_name(
                        bids_meta, this_prefix_basename, echo_times_lst
                    )

                if is_complex:
                    this_prefix_basename = update_complex_name(
                        bids_meta, this_prefix_basename
                    )

                if is_uncombined:
                    this_prefix_basename = update_uncombined_name(
                        bids_meta, this_prefix_basename, channel_names_lst
                    )

            # Fallback option:
            # If we have failed to modify this_prefix_basename, because it didn't fall
            #   into any of the options above, just add the suffix at the end:
            if this_prefix_basename == prefix_basename:
                this_prefix_basename += suffix

            # Finally, form the outname by stitching the directory and outtype:
            outname = op.join(prefix_dirname, this_prefix_basename)
            outfile = outname + "." + outtype

            # Write the files needed:
            safe_movefile(fl, outfile, overwrite)
            if bids_file:
                outname_bids_file = "%s.json" % (outname)
                safe_movefile(bids_file, outname_bids_file, overwrite)
                bids_outfiles.append(outname_bids_file)

    # res_files is not a list
    else:
        outname = "{}.{}".format(prefix, outtype)
        safe_movefile(res_files, outname, overwrite)
        if isdefined(res.outputs.bids):
            try:
                safe_movefile(res.outputs.bids, outname_bids, overwrite)
                bids_outfiles.append(outname_bids)
            except TypeError:  ##catch lists
                raise TypeError("Multiple BIDS sidecars detected.")
    return bids_outfiles


def add_taskname_to_infofile(infofiles: str | list[str]) -> None:
    """Add the "TaskName" field to json files with _task- entity in the name.

    Note: _task- entity could be present not only in functional data
    but in many other modalities now.

    Parameters
    ----------
    infofiles: list or str
        json filenames or a single filename.
    """

    # in case they pass a string with a path:
    if isinstance(infofiles, str):
        infofiles = [infofiles]

    for infofile in infofiles:
        meta_info = load_json(infofile)
        m = re.search(r"(?<=_task-)\w+", op.basename(infofile))
        if m:
            meta_info["TaskName"] = m.group(0).split("_")[0]
        else:
            # leave it to bids-validator to validate/inform about presence
            # of required entities/fields.
            continue

        # write to outfile
        save_json(infofile, meta_info)


def bvals_are_zero(bval_file: str | list) -> bool:
    """Checks if all entries in a bvals file are zero (or 5, for Siemens files).

    Parameters
    ----------
    bval_file : str
      file with the bvals

    Returns
    -------
    True if all are all 0 or 5; False otherwise.
    """

    # GE hyperband multi-echo containing diffusion info
    if isinstance(bval_file, list):
        return all(map(bvals_are_zero, bval_file))

    with open(bval_file) as f:
        bvals = f.read().split()

    bvals_unique = set(float(b) for b in bvals)
    return bvals_unique == {0.0} or bvals_unique == {5.0}



"""Handle BIDS specific operations"""

from __future__ import annotations

__docformat__ = "numpy"

from collections import OrderedDict
import csv
import errno
from glob import glob
import hashlib
import logging
import os
import os.path as op
from pathlib import Path
import re
from typing import Any, Optional
import warnings

import numpy as np
import pydicom as dcm

from . import __version__, dicoms
from .parser import find_files
from .utils import (
    create_file_if_missing,
    is_readonly,
    json_dumps,
    load_json,
    remove_prefix,
    remove_suffix,
    save_json,
    set_readonly,
    strptime_micr,
    update_json,
)

lgr = logging.getLogger(__name__)

# Fields to be populated in _scans files. Order matters
SCANS_FILE_FIELDS = OrderedDict(
    [
        ("filename", OrderedDict([("Description", "Name of the nifti file")])),
        (
            "acq_time",
            OrderedDict(
                [
                    ("LongName", "Acquisition time"),
                    ("Description", "Acquisition time of the particular scan"),
                ]
            ),
        ),
        ("operator", OrderedDict([("Description", "Name of the operator")])),
        (
            "randstr",
            OrderedDict(
                [("LongName", "Random string"), ("Description", "md5 hash of UIDs")]
            ),
        ),
    ]
)

#: JSON Key where we will embed our version in the newly produced .json files
HEUDICONV_VERSION_JSON_KEY = "HeudiconvVersion"


class BIDSError(Exception):
    pass


BIDS_VERSION = "1.8.0"

# List defining allowed parameter matching for fmap assignment:
SHIM_KEY = "ShimSetting"
AllowedFmapParameterMatching = [
    "Shims",
    "ImagingVolume",
    "ModalityAcquisitionLabel",
    "CustomAcquisitionLabel",
    "Force",
]
# Key info returned by get_key_info_for_fmap_assignment when
# matching_parameter = "Force"
KeyInfoForForce = "Forced"
# List defining allowed criteria to assign a given fmap to a non-fmap run
# among the different fmaps with matching parameters:
AllowedCriteriaForFmapAssignment = [
    "First",
    "Closest",
]


def maybe_na(val: Any) -> str:
    """Return 'n/a' if non-None value represented as str is not empty

    Primarily for the consistent use of lower case 'n/a' so 'N/A' and 'NA'
    are also treated as 'n/a'
    """
    if val is not None:
        valstr = str(val).strip()
        return "n/a" if (not valstr or valstr in ("N/A", "NA")) else valstr
    else:
        return "n/a"


def treat_age(age: str | float | None) -> str | None:
    """Age might encounter 'Y' suffix or be a float"""
    if age is None:
        return None  # might be converted to N/A by maybe_na
    agestr = str(age)
    if agestr.endswith("M"):
        agestr = agestr.rstrip("M")
        ageflt = float(agestr) / 12
        agestr = ("%.2f" if ageflt != int(ageflt) else "%d") % ageflt
    else:
        agestr = agestr.rstrip("Y")
    if agestr:
        # strip all leading 0s but allow to scan a newborn (age 0Y)
        agestr = "0" if not agestr.lstrip("0") else agestr.lstrip("0")
        if agestr.startswith("."):
            # we had float point value, let's prepend 0
            agestr = "0" + agestr
    return agestr


def populate_bids_templates(
    path: str, defaults: Optional[dict[str, Any]] = None
) -> None:
    """Premake BIDS text files with templates"""

    lgr.info("Populating template files under %s", path)
    descriptor = op.join(path, "dataset_description.json")
    if defaults is None:
        defaults = {}
    if not op.lexists(descriptor):
        save_json(
            descriptor,
            OrderedDict(
                [
                    ("Name", "TODO: name of the dataset"),
                    ("BIDSVersion", BIDS_VERSION),
                    (
                        "License",
                        defaults.get(
                            "License",
                            "TODO: choose a license, e.g. PDDL "
                            "(http://opendatacommons.org/licenses/pddl/)",
                        ),
                    ),
                    (
                        "Authors",
                        defaults.get(
                            "Authors", ["TODO:", "First1 Last1", "First2 Last2", "..."]
                        ),
                    ),
                    (
                        "Acknowledgements",
                        defaults.get(
                            "Acknowledgements", "TODO: whom you want to acknowledge"
                        ),
                    ),
                    (
                        "HowToAcknowledge",
                        "TODO: describe how to acknowledge -- either cite a "
                        "corresponding paper, or just in acknowledgement "
                        "section",
                    ),
                    ("Funding", ["TODO", "GRANT #1", "GRANT #2"]),
                    ("ReferencesAndLinks", ["TODO", "List of papers or websites"]),
                    ("DatasetDOI", "TODO: eventually a DOI for the dataset"),
                ]
            ),
        )
    sourcedata_README = op.join(path, "sourcedata", "README")
    if op.exists(op.dirname(sourcedata_README)):
        create_file_if_missing(
            sourcedata_README,
            (
                "TODO: Provide description about source data, e.g. \n"
                "Directory below contains DICOMS compressed into tarballs per "
                "each sequence, replicating directory hierarchy of the BIDS dataset"
                " itself."
            ),
        )
    create_file_if_missing(
        op.join(path, "CHANGES"),
        "0.0.1  Initial data acquired\n"
        "TODOs:\n\t- verify and possibly extend information in participants.tsv"
        " (see for example http://datasets.datalad.org/?dir=/openfmri/ds000208)"
        "\n\t- fill out dataset_description.json, README, sourcedata/README"
        " (if present)\n\t- provide _events.tsv file for each _bold.nii.gz with"
        " onsets of events (see  '8.5 Task events'  of BIDS specification)",
    )
    create_file_if_missing(
        op.join(path, "README"),
        "TODO: Provide description for the dataset -- basic details about the "
        "study, possibly pointing to pre-registration (if public or embargoed)",
    )
    create_file_if_missing(
        op.join(path, "scans.json"), json_dumps(SCANS_FILE_FIELDS, sort_keys=False)
    )
    create_file_if_missing(op.join(path, ".bidsignore"), ".duecredit.p")
    if op.lexists(op.join(path, ".git")):
        create_file_if_missing(op.join(path, ".gitignore"), ".duecredit.p")

    populate_aggregated_jsons(path)


def populate_aggregated_jsons(path: str) -> None:
    """Aggregate across the entire BIDS dataset ``.json``\\s into top level ``.json``\\s

    Top level .json files would contain only the fields which are
    common to all ``subject[/session]/type/*_modality.json``\\s.

    ATM aggregating only for ``*_task*_bold.json`` files. Only the task- and
    OPTIONAL _acq- field is retained within the aggregated filename.  The other
    BIDS _key-value pairs are "aggregated over".

    Parameters
    ----------
    path: str
      Path to the top of the BIDS dataset
    """
    # TODO: collect all task- .json files for func files to
    tasks = {}
    # way too many -- let's just collect all which are the same!
    # FIELDS_TO_TRACK = {'RepetitionTime', 'FlipAngle', 'EchoTime',
    #                    'Manufacturer', 'SliceTiming', ''}
    for fpath in find_files(
        r".*_task-.*\_bold\.json",
        topdir=glob(op.join(path, "sub-*")),
        exclude_vcs=True,
        exclude=r"/\.(datalad|heudiconv)/",
    ):
        #
        # According to BIDS spec I think both _task AND _acq (may be more?
        # _rec, _dir, ...?) should be retained?
        # TODO: if we are to fix it, then old ones (without _acq) should be
        # removed first
        task = re.sub(r".*_(task-[^_\.]*(_acq-[^_\.]*)?)_.*", r"\1", fpath)
        json_ = load_json(fpath, retry=100)
        if task not in tasks:
            tasks[task] = json_
        else:
            rec = tasks[task]
            # let's retain only those fields which have the same value
            for field in sorted(rec):
                if field not in json_ or json_[field] != rec[field]:
                    del rec[field]
        # create a stub onsets file for each one of those
        suf = "_bold.json"
        assert fpath.endswith(suf)
        # specify the name of the '_events.tsv' file:
        if "_echo-" in fpath:
            # multi-echo sequence: bids (1.1.0) specifies just one '_events.tsv'
            #   file, common for all echoes.  The name will not include _echo-.
            # TODO: RF to use re.match for better readability/robustness
            # So, find out the echo number:
            fpath_split = fpath.split("_echo-", 1)  # split fpath using '_echo-'
            fpath_split_2 = fpath_split[1].split(
                "_", 1
            )  # split the second part of fpath_split using '_'
            echoNo = fpath_split_2[0]  # get echo number
            if echoNo == "1":
                if len(fpath_split_2) != 2:
                    raise ValueError("Found no trailer after _echo-")
                # we modify fpath to exclude '_echo-' + echoNo:
                fpath = fpath_split[0] + "_" + fpath_split_2[1]
            else:
                # for echoNo greater than 1, don't create the events file, so go to
                #   the next for loop iteration:
                continue

        events_file = remove_suffix(fpath, suf) + "_events.tsv"
        # do not touch any existing thing, it may be precious
        if not op.lexists(events_file):
            lgr.debug("Generating %s", events_file)
            with open(events_file, "w") as fp:
                fp.write(
                    "onset\tduration\ttrial_type\tresponse_time\tstim_file"
                    "\tTODO -- fill in rows and add more tab-separated "
                    "columns if desired"
                )
    # extract tasks files stubs
    for task_acq, fields in tasks.items():
        task_file = op.join(path, task_acq + "_bold.json")
        # Since we are pulling all unique fields we have to possibly
        # rewrite this file to guarantee consistency.
        # See https://github.com/nipy/heudiconv/issues/277 for a usecase/bug
        # when we didn't touch existing one.
        # But the fields we enter (TaskName and CogAtlasID) might need need
        # to be populated from the file if it already exists
        placeholders = {
            "TaskName": (
                "TODO: full task name for %s" % task_acq.split("_")[0].split("-")[1]
            ),
            "CogAtlasID": "http://www.cognitiveatlas.org/task/id/TODO",
        }
        if op.lexists(task_file):
            j = load_json(task_file, retry=100)
            # Retain possibly modified placeholder fields
            for f in placeholders:
                if f in j:
                    placeholders[f] = j[f]
            act = "Regenerating"
        else:
            act = "Generating"
        lgr.debug("%s %s", act, task_file)
        fields.update(placeholders)
        save_json(task_file, fields, sort_keys=True, pretty=True)


def tuneup_bids_json_files(json_files: list[str]) -> None:
    """Given a list of BIDS .json files, e.g."""
    if not json_files:
        return
    # Harmonize generic .json formatting
    for jsonfile in json_files:
        json_ = load_json(jsonfile)
        # sanitize!
        for f1 in ["Acquisition", "Study", "Series"]:
            for f2 in ["DateTime", "Date"]:
                json_.pop(f1 + f2, None)
        # TODO:  should actually be placed into series file which must
        #        go under annex (not under git) and marked as sensitive
        # MG - Might want to replace with flag for data sensitivity
        # related - https://github.com/nipy/heudiconv/issues/92
        if "Date" in str(json_):
            # Let's hope no word 'Date' comes within a study name or smth like
            # that
            raise ValueError("There must be no dates in .json sidecar")
        # Those files should not have our version field already - should have been
        # freshly produced
        assert HEUDICONV_VERSION_JSON_KEY not in json_
        json_[HEUDICONV_VERSION_JSON_KEY] = str(__version__)
        save_json(jsonfile, json_)

    # Load the beast
    seqtype = op.basename(op.dirname(jsonfile))

    # MG - want to expand this for other _epi
    # possibly add IntendedFor automatically as well?
    if seqtype == "fmap":
        json_basename = "_".join(jsonfile.split("_")[:-1])
        # if we got by now all needed .json files -- we can fix them up
        # unfortunately order of "items" is not guaranteed atm
        json_phasediffname = json_basename + "_phasediff.json"
        json_mag = json_basename + "_magnitude*.json"
        if op.exists(json_phasediffname) and len(glob(json_mag)) >= 1:
            json_ = load_json(json_phasediffname)
            # TODO: we might want to reorder them since ATM
            # the one for shorter TE is the 2nd one!
            # For now just save truthfully by loading magnitude files
            lgr.debug("Placing EchoTime fields into phasediff file")
            for i in 1, 2:
                try:
                    json_["EchoTime%d" % i] = load_json(
                        json_basename + "_magnitude%d.json" % i
                    )["EchoTime"]
                except IOError as exc:
                    lgr.error("Failed to open magnitude file: %s", exc)
            # might have been made R/O already, but if not -- it will be set
            # only later in the pipeline, so we must not make it read-only yet
            was_readonly = is_readonly(json_phasediffname)
            if was_readonly:
                set_readonly(json_phasediffname, False)
            save_json(json_phasediffname, json_)
            if was_readonly:
                set_readonly(json_phasediffname)


def add_participant_record(
    studydir: str, subject: str, age: str | None, sex: str | None
) -> None:
    participants_tsv = op.join(studydir, "participants.tsv")
    participant_id = "sub-%s" % subject

    if not create_file_if_missing(
        participants_tsv, "\t".join(["participant_id", "age", "sex", "group"]) + "\n"
    ):
        # check if may be subject record already exists
        with open(participants_tsv) as f:
            f.readline()
            known_subjects = {ln.split("\t")[0] for ln in f.readlines()}
        if participant_id in known_subjects:
            return
    else:
        # Populate particpants.json (an optional file to describe column names in
        # participant.tsv). This auto generation will make BIDS-validator happy.
        participants_json = op.join(studydir, "participants.json")
        if not op.lexists(participants_json):
            save_json(
                participants_json,
                OrderedDict(
                    [
                        (
                            "participant_id",
                            OrderedDict([("Description", "Participant identifier")]),
                        ),
                        (
                            "age",
                            OrderedDict(
                                [
                                    (
                                        "Description",
                                        "Age in years (TODO - verify) as in the initial"
                                        " session, might not be correct for other sessions",
                                    )
                                ]
                            ),
                        ),
                        (
                            "sex",
                            OrderedDict(
                                [
                                    (
                                        "Description",
                                        "self-rated by participant, M for male/F for "
                                        "female (TODO: verify)",
                                    )
                                ]
                            ),
                        ),
                        (
                            "group",
                            OrderedDict(
                                [
                                    (
                                        "Description",
                                        "(TODO: adjust - by default everyone is in "
                                        "control group)",
                                    )
                                ]
                            ),
                        ),
                    ]
                ),
                sort_keys=False,
            )

    # Add a new participant
    with open(participants_tsv, "a") as f:
        f.write(
            "\t".join(
                map(
                    str,
                    [
                        participant_id,
                        maybe_na(treat_age(age)),
                        maybe_na(sex),
                        "control",
                    ],
                )
            )
            + "\n"
        )


def find_subj_ses(f_name: str) -> tuple[Optional[str], Optional[str]]:
    """Given a path to the bids formatted filename parse out subject/session"""
    # we will allow the match at either directories or within filename
    # assuming that bids layout is "correct"
    regex = re.compile("sub-(?P<subj>[a-zA-Z0-9]*)([/_]ses-(?P<ses>[a-zA-Z0-9]*))?")
    regex_res = regex.search(f_name)
    res = regex_res.groupdict() if regex_res else {}
    return res.get("subj", None), res.get("ses", None)


def save_scans_key(
    item: tuple[str, tuple[str, ...], list[str]], bids_files: list[str]
) -> None:
    """
    Parameters
    ----------
    item:
    bids_files: list of str

    Returns
    -------

    """
    rows = {}
    assert bids_files, "we do expect some files since it was called"
    # we will need to deduce subject and session from the bids_filename
    # and if there is a conflict, we would just blow since this function
    # should be invoked only on a result of a single item conversion as far
    # as I see it, so should have the same subject/session
    subj: Optional[str] = None
    ses: Optional[str] = None
    for bids_file in bids_files:
        # get filenames
        f_name = "/".join(bids_file.split("/")[-2:])
        f_name = f_name.replace("json", "nii.gz")
        rows[f_name] = get_formatted_scans_key_row(item[-1][0])
        subj_, ses_ = find_subj_ses(f_name)
        if not subj_:
            lgr.warning(
                "Failed to detect fulfilled BIDS layout.  "
                "No scans.tsv file(s) will be produced for %s",
                ", ".join(bids_files),
            )
            return
        if subj and subj_ != subj:
            raise ValueError(
                "We found before subject %s but now deduced %s from %s"
                % (subj, subj_, f_name)
            )
        subj = subj_
        if ses and ses_ != ses:
            raise ValueError(
                "We found before session %s but now deduced %s from %s"
                % (ses, ses_, f_name)
            )
        ses = ses_
    # where should we store it?
    output_dir = op.dirname(op.dirname(bids_file))
    # save
    ses = "_ses-%s" % ses if ses else ""
    add_rows_to_scans_keys_file(
        op.join(output_dir, "sub-{0}{1}_scans.tsv".format(subj, ses)), rows
    )


def add_rows_to_scans_keys_file(fn: str, newrows: dict[str, list[str]]) -> None:
    """Add new rows to the _scans file.

    Parameters
    ----------
    fn: str
      filename
    newrows: dict
      extra rows to add (acquisition time, referring physician, random string)
    """
    if op.lexists(fn):
        with open(fn, "r") as csvfile:
            reader = csv.reader(csvfile, delimiter="\t")
            existing_rows = [row for row in reader]
        # skip header
        fnames2info = {row[0]: row[1:] for row in existing_rows[1:]}

        newrows_key = newrows.keys()
        newrows_toadd = list(set(newrows_key) - set(fnames2info.keys()))
        for key_toadd in newrows_toadd:
            fnames2info[key_toadd] = newrows[key_toadd]
        # remove
        os.unlink(fn)
    else:
        fnames2info = newrows

    header = list(SCANS_FILE_FIELDS.keys())
    # prepare all the data rows
    data_rows = [[k] + v for k, v in fnames2info.items()]
    # sort by the date/filename
    try:
        data_rows_sorted = sorted(data_rows, key=lambda x: (x[1], x[0]))
    except TypeError as exc:
        lgr.warning("Sorting scans by date failed: %s", str(exc))
        data_rows_sorted = sorted(data_rows)
    # save
    with open(fn, "a") as csvfile:
        writer = csv.writer(csvfile, delimiter="\t")
        writer.writerows([header] + data_rows_sorted)


def get_formatted_scans_key_row(dcm_fn: str | Path) -> list[str]:
    """
    Parameters
    ----------
    dcm_fn: str

    Returns
    -------
    row: list
        [ISO acquisition time, performing physician name, random string]

    """
    dcm_data = dcm.dcmread(dcm_fn, stop_before_pixels=True, force=True)
    # we need to store filenames and acquisition datetimes
    acq_datetime = dicoms.get_datetime_from_dcm(dcm_data=dcm_data)
    # add random string
    # But let's make it reproducible by using all UIDs
    # (might change across versions?)
    randcontent = "".join(
        [getattr(dcm_data, f) or "" for f in sorted(dir(dcm_data)) if f.endswith("UID")]
    )
    randstr = hashlib.md5(randcontent.encode()).hexdigest()[:8]
    try:
        perfphys = dcm_data.PerformingPhysicianName
    except AttributeError:
        perfphys = ""
    row = [acq_datetime.isoformat() if acq_datetime else "", perfphys, randstr]
    # empty entries should be 'n/a'
    # https://github.com/dartmouth-pbs/heudiconv/issues/32
    row = ["n/a" if not str(e) else e for e in row]
    return row


def convert_sid_bids(subject_id: str) -> str:
    """Shim for stripping any non-BIDS compliant characters within subject_id

    Parameters
    ----------
    subject_id : string

    Returns
    -------
    sid : string
        New subject ID
    """
    warnings.warn(
        "convert_sid_bids() is deprecated, please use sanitize_label() instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return sanitize_label(subject_id)


def get_shim_setting(json_file: str) -> Any:
    """
    Gets the "ShimSetting" field from a json_file.
    If no "ShimSetting" present, return error

    Parameters
    ----------
    json_file : str

    Returns
    -------
    str with "ShimSetting" value
    """
    data = load_json(json_file)
    try:
        shims = data[SHIM_KEY]
    except KeyError:
        lgr.error(
            'File %s does not have "%s". '
            'Please use a different "matching_parameters" in your heuristic file',
            json_file,
            SHIM_KEY,
        )
        raise
    return shims


def find_fmap_groups(fmap_dir: str) -> dict[str, list[str]]:
    """
    Finds the different fmap groups in a fmap directory.
    By groups here we mean fmaps that are intended to go together
    (with reversed PE polarity, magnitude/phase, etc.)

    Parameters
    ----------
    fmap_dir : str
        path to the session folder (or to the subject folder, if there are no
        sessions).

    Returns
    -------
    fmap_groups : dict
        key: prefix common to the group (e.g. no "dir" entity, "_phase"/"_magnitude", ...)
        value: list of all fmap paths in the group
    """
    if op.basename(fmap_dir) != "fmap":
        lgr.error("%s is not a fieldmap folder", fmap_dir)

    # Get a list of all fmap json files in the session:
    fmap_jsons = sorted(glob(op.join(fmap_dir, "*.json")))

    # RegEx to remove fmap-specific substrings from fmap file names
    # "_phase[1,2]", "_magnitude[1,2]", "_phasediff", "_dir-<label>", ...
    fmap_regex = re.compile(
        "(_dir-[0-9,a-z,A-Z]*)*"  # for pepolar case
        "(_phase[12])*"  # for phase images
        "(_phasediff)*"  # for phasediff images
        "(_magnitude[12])*"  # for magnitude images
        "(_fieldmap)*"  # for actual fieldmap images
    )

    # Find the unique prefixes ('splitext' removes the extension):
    prefixes = sorted(
        set(
            fmap_regex.sub("", remove_suffix(op.basename(fm), ".json"))
            for fm in fmap_jsons
        )
    )
    return {
        k: [
            fm
            for fm in fmap_jsons
            if fmap_regex.sub("", remove_suffix(op.basename(fm), ".json")) == k
        ]
        for k in prefixes
    }


def get_key_info_for_fmap_assignment(
    json_file: str, matching_parameter: str
) -> list[Any]:
    """
    Gets key information needed to assign fmaps to other modalities.
    (Note: It is the responsibility of the calling function to make sure
    the arguments are OK)

    Parameters
    ----------
    json_file : str
        path to the json file
    matching_parameter : str in AllowedFmapParameterMatching
        matching_parameter that will be used to match runs

    Returns
    -------
    key_info : list
        part of the json file that will need to match between the fmap and
        the other image
    """
    if not op.exists(json_file):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), json_file)

    # loop through the possible criteria and extract the info needed
    if matching_parameter == "Shims":
        key_info = [get_shim_setting(json_file)]
    elif matching_parameter == "ImagingVolume":
        from nibabel import load as nb_load
        from nibabel.nifti1 import Nifti1Header

        nifti_files = glob(remove_suffix(json_file, ".json") + ".nii*")
        assert len(nifti_files) == 1
        nifti_file = nifti_files[0]
        nifti_header = nb_load(nifti_file).header
        assert isinstance(nifti_header, Nifti1Header)
        key_info = [nifti_header.get_best_affine(), nifti_header.get_data_shape()[:3]]
    elif matching_parameter == "ModalityAcquisitionLabel":
        # Check the acq label for the fmap and the modality for others:
        modality = op.basename(op.dirname(json_file))
        if modality == "fmap":
            # extract the <acq> entity:
            acq_label = BIDSFile.parse(op.basename(json_file))["acq"]
            assert acq_label is not None
            if any(s in acq_label.lower() for s in ["fmri", "bold", "func"]):
                key_info = ["func"]
            elif any(s in acq_label.lower() for s in ["diff", "dwi"]):
                key_info = ["dwi"]
            elif any(s in acq_label.lower() for s in ["anat", "struct"]):
                key_info = ["anat"]
        else:
            key_info = [modality]
    elif matching_parameter == "CustomAcquisitionLabel":
        modality = op.basename(op.dirname(json_file))
        if modality == "func":
            # extract the <task> entity:
            custom_label = BIDSFile.parse(op.basename(json_file))["task"]
        else:
            # extract the <acq> entity:
            custom_label = BIDSFile.parse(op.basename(json_file))["acq"]
        # Get the custom acquisition label, acq_label is None if no custom field found
        key_info = [custom_label]
    elif matching_parameter == "Force":
        # We want to force the matching, so just return some string
        # regardless of the image
        key_info = [KeyInfoForForce]
    else:
        # fallback:
        key_info = []

    return key_info


def find_compatible_fmaps_for_run(
    json_file: str, fmap_groups: dict[str, list[str]], matching_parameters: list[str]
) -> dict[str, list[str]]:
    """
    Finds compatible fmaps for a given run, for populate_intended_for.
    (Note: It is the responsibility of the calling function to make sure
    the arguments are OK)

    Parameters
    ----------
    json_file : str
        path to the json file
    fmap_groups : dict
        key: prefix common to the group
        value: list of all fmap paths in the group
    matching_parameters : list of str from AllowedFmapParameterMatching
        matching_parameters that will be used to match runs

    Returns
    -------
    compatible_fmap_groups : dict
        Subset of the fmap_groups which match json_file, according
        to the matching_parameters.
        key: prefix common to the group
        value: list of all fmap paths in the group
    """
    lgr.debug("Looking for fmaps for %s", json_file)
    json_info = {}
    for param in matching_parameters:
        json_info[param] = get_key_info_for_fmap_assignment(json_file, param)

    compatible_fmap_groups = {}
    for fm_key, fm_group in fmap_groups.items():
        # check the key_info (for all parameters) for one (the first) of
        # the fmaps in the group:
        compatible = False
        for param in matching_parameters:
            json_info_1st_item = json_info[param][0]
            fm_info = get_key_info_for_fmap_assignment(fm_group[0], param)
            # for the case in which key_info is a list of strings:
            if isinstance(json_info_1st_item, str):
                compatible = json_info[param] == fm_info
            # for the case when no key info was found (e.g. "acq" field does not exist)
            elif json_info_1st_item is None:
                compatible = False
            else:
                # allow for tiny differences between the affines etc
                compatible = all(
                    np.allclose(x, y) for x, y in zip(json_info[param], fm_info)
                )
            if not compatible:
                continue  # don't bother checking more params
        if compatible:
            compatible_fmap_groups[fm_key] = fm_group

    return compatible_fmap_groups


def find_compatible_fmaps_for_session(
    path_to_bids_session: str, matching_parameters: list[str]
) -> Optional[dict[str, dict[str, list[str]]]]:
    """
    Finds compatible fmaps for all non-fmap runs in a session.
    (Note: It is the responsibility of the calling function to make sure
    the arguments are OK)

    Parameters
    ----------
    path_to_bids_session : str
        path to the session folder (or to the subject folder, if there are no
        sessions).
    matching_parameters : list of str from AllowedFmapParameterMatching
        matching_parameters that will be used to match runs

    Returns
    -------
    compatible_fmap : dict
        Dict of compatible_fmaps_groups (values) for each non-fmap run (keys)
    """
    lgr.debug("Looking for fmaps for session: %s", path_to_bids_session)

    # Resolve path (eliminate '..')
    path_to_bids_session = op.abspath(path_to_bids_session)

    # find the different groups of fmaps:
    fmap_dir = op.join(path_to_bids_session, "fmap")
    if not op.exists(fmap_dir):
        lgr.warning(
            "We cannot add the IntendedFor field: no fmap/ in %s", path_to_bids_session
        )
        return None
    fmap_groups = find_fmap_groups(fmap_dir)

    # Get a set with all non-fmap json files in the session (exclude SBRef files).
    session_jsons = [
        j
        for j in glob(op.join(path_to_bids_session, "*/*.json"))
        if not (
            op.basename(op.dirname(j)) == "fmap"
            or remove_suffix(j, ".json").endswith("_sbref")
        )
    ]

    # Loop through session_jsons and find the compatible fmap_groups for each
    compatible_fmaps = {
        j: find_compatible_fmaps_for_run(j, fmap_groups, matching_parameters)
        for j in session_jsons
    }
    return compatible_fmaps


def select_fmap_from_compatible_groups(
    json_file: str, compatible_fmap_groups: dict[str, list[str]], criterion: str
) -> Optional[str]:
    """
    Selects the fmap that will be used to correct for distortions in json_file
    from the compatible fmap_groups list, based on the given criterion
    (Note: It is the responsibility of the calling function to make sure
    the arguments are OK)

    Parameters
    ----------
    json_file : str
        path to the json file
    compatible_fmap_groups : dict
        fmap_groups that are compatible with the specific json_file
    criterion : str in ['First', 'Closest']
        matching_parameters that will be used to decide which fmap to use

    Returns
    -------
    selected_fmap_key : str
        key from the compatible_fmap_groups for the selected fmap group
    """
    if len(compatible_fmap_groups) == 0:
        return None
    # if compatible_fmap_groups has only one entry, that's it:
    elif len(compatible_fmap_groups) == 1:
        return list(compatible_fmap_groups.keys())[0]

    # get the modality folders, then session folder:
    modality_folders = set(
        op.dirname(fmap) for v in compatible_fmap_groups.values() for fmap in v
    )  # there should be only one value, ending in 'fmap'
    sess_folders = set(op.dirname(k) for k in modality_folders)
    if len(sess_folders) > 1:
        # for now, we only deal with single sessions:
        raise RuntimeError
    # if we made it here, we have only one session:
    sess_folder = list(sess_folders)[0]

    # get acquisition times from '_scans.tsv':
    try:
        scans_tsv = glob(op.join(sess_folder, "*_scans.tsv"))[0]
    except IndexError:
        raise FileNotFoundError("No '*_scans' file found for session %s" % sess_folder)
    with open(scans_tsv) as f:
        # read the contents, splitting by lines and by tab separators:
        scans_tsv_content = [line.split("\t") for line in f.read().splitlines()]
    # get column indices for filename and acq_time from the first line:
    (fname_idx, time_idx) = (
        scans_tsv_content[0].index(k) for k in ["filename", "acq_time"]
    )
    acq_times = {line[fname_idx]: line[time_idx] for line in scans_tsv_content[1:]}
    # acq_times for the compatible fmaps:
    acq_times_fmaps = {
        k: acq_times[
            # remove session folder and '.json', add '.nii.gz':
            remove_suffix(remove_prefix(v[0], sess_folder + op.sep), ".json")
            + ".nii.gz"
        ]
        for k, v in compatible_fmap_groups.items()
    }

    if criterion == "First":
        # find the first acquired fmap_group from the compatible_fmap_groups:
        first_acq_time = sorted(acq_times_fmaps.values())[0]
        selected_fmap_key = [
            k for k, v in acq_times_fmaps.items() if v == first_acq_time
        ][0]
    elif criterion == "Closest":
        json_acq_time = strptime_micr(
            acq_times[
                # remove session folder and '.json', add '.nii.gz':
                remove_suffix(remove_prefix(json_file, sess_folder + op.sep), ".json")
                + ".nii.gz"
            ],
            "%Y-%m-%dT%H:%M:%S[.%f]",
        )
        # differences in acquisition time (abs value):
        diff_fmaps_acq_times = {
            k: abs(strptime_micr(v, "%Y-%m-%dT%H:%M:%S[.%f]") - json_acq_time)
            for k, v in acq_times_fmaps.items()
        }
        min_diff_acq_times = sorted(diff_fmaps_acq_times.values())[0]
        selected_fmap_key = [
            k for k, v in diff_fmaps_acq_times.items() if v == min_diff_acq_times
        ][0]
    else:
        raise ValueError(f"Invalid 'criterion' value: {criterion!r}")

    return selected_fmap_key


def populate_intended_for(
    path_to_bids_session: str, matching_parameters: str | list[str], criterion: str
) -> None:
    """
    Adds the 'IntendedFor' field to the fmap .json files in a session folder.
    It goes through the session folders and for every json file, it finds
    compatible_fmaps: fmaps that have the same matching_parameters as the json
    file (e.g., same 'Shims').

    If there are more than one compatible_fmaps, it will use the criterion
    specified by the user (default: 'Closest' in time).

    Because fmaps come in groups (with reversed PE polarity, or magnitude/
    phase), we work with fmap_groups.

    Parameters
    ----------
    path_to_bids_session : str
        path to the session folder (or to the subject folder, if there are no
        sessions).
    matching_parameters : list of str from AllowedFmapParameterMatching
        matching_parameters that will be used to match runs
    criterion : str in ['First', 'Closest']
        matching_parameters that will be used to decide which of the matching
        fmaps to use
    """

    if not isinstance(matching_parameters, list):
        assert isinstance(matching_parameters, str), (
            "matching_parameters must be a str or a list, got %s" % matching_parameters
        )
        matching_parameters = [matching_parameters]
    for param in matching_parameters:
        if param not in AllowedFmapParameterMatching:
            raise ValueError("Fmap matching_parameter %s not allowed." % param)
    if criterion not in AllowedCriteriaForFmapAssignment:
        raise ValueError("Fmap assignment criterion '%s' not allowed." % criterion)

    lgr.info('Adding "IntendedFor" to the fieldmaps in %s.', path_to_bids_session)

    # Resolve path (eliminate '..')
    path_to_bids_session = op.abspath(path_to_bids_session)

    # Get the subject folder (if "path_to_bids_session" includes the session,
    # remove it). "IntendedFor" paths will be relative to it.
    if op.basename(path_to_bids_session).startswith("ses-"):
        subj_folder = op.dirname(path_to_bids_session)
    else:
        subj_folder = path_to_bids_session

    fmap_dir = op.join(path_to_bids_session, "fmap")
    if not op.exists(fmap_dir):
        lgr.warning(
            "We cannot add the IntendedFor field: no fmap/ in %s", path_to_bids_session
        )
        return

    compatible_fmaps = find_compatible_fmaps_for_session(
        path_to_bids_session, matching_parameters=matching_parameters
    )
    assert compatible_fmaps is not None
    selected_fmaps = {}
    for json_file, fmap_groups in compatible_fmaps.items():
        if not op.dirname(json_file).endswith("fmap"):
            selected_fmaps[json_file] = select_fmap_from_compatible_groups(
                json_file, fmap_groups, criterion=criterion
            )

    # Loop through all the unique fmap_groups in compatible_fmaps:
    unique_fmap_groups = {}
    for cf in compatible_fmaps.values():
        for key, values in cf.items():
            if key not in unique_fmap_groups:
                unique_fmap_groups[key] = values

    for fmap_group in unique_fmap_groups:
        intended_for = []
        for json_file, selected_fmap_group in selected_fmaps.items():
            if selected_fmap_group and (fmap_group in selected_fmap_group):
                intended_for.append(
                    op.relpath(
                        remove_suffix(json_file, ".json") + ".nii.gz", start=subj_folder
                    )
                )
        if intended_for:
            intended_for = sorted(str(f) for f in intended_for)
            # Add this intended_for to all fmap files in the fmap_group:
            for fm_json in unique_fmap_groups[fmap_group]:
                update_json(fm_json, {"IntendedFor": intended_for}, pretty=True)


class BIDSFile:
    """as defined in https://bids-specification.readthedocs.io/en/stable/99-appendices/04-entity-table.html
    which might soon become machine readable
    order matters
    """

    _known_entities = [
        "sub",
        "ses",
        "task",
        "acq",
        "ce",
        "rec",
        "dir",
        "run",
        "mod",
        "echo",
        "flip",
        "inv",
        "mt",
        "part",
        "recording",
    ]

    def __init__(
        self, entities: dict[str, str], suffix: str, extension: Optional[str]
    ) -> None:
        self._entities = entities
        self._suffix = suffix
        self._extension = extension

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        if (
            all([other[k] == v for k, v in self._entities.items()])
            and self.extension == other.extension
            and self.suffix == other.suffix
        ):
            return True
        else:
            return False

    @classmethod
    def parse(cls, filename: str) -> BIDSFile:
        """Parse the filename for BIDS entities, suffix and extension"""
        # use re.findall to find all lower-case-letters + '-' + alphanumeric + '_' pairs:
        entities_list = re.findall("([a-z]+)-([a-zA-Z0-9]+)[_]*", filename)
        # keep only those in the _known_entities list:
        entities = {k: v for k, v in entities_list if k in BIDSFile._known_entities}
        # get whatever comes after the last key-value pair, and remove any '_' that
        # might come in front:
        ending = filename.split("-".join(entities_list[-1]))[-1]
        ending = remove_prefix(ending, "_")
        # the first dot ('.') separates the suffix from the extension:
        if "." in ending:
            suffix, extension = ending.split(".", 1)
        else:
            suffix, extension = ending, None
        return BIDSFile(entities, suffix, extension)

    def __str__(self) -> str:
        """reconstitute in a legit BIDS filename using the order from entity table"""
        if "sub" not in self._entities:
            raise ValueError("The 'sub-' entity is mandatory")
        # reconstitute the ending for the filename:
        suffix = "_" + self.suffix if self.suffix else ""
        extension = "." + self.extension if self.extension else ""
        return (
            "_".join(
                [
                    "-".join([e, self._entities[e]])
                    for e in self._known_entities
                    if e in self._entities
                ]
            )
            + suffix
            + extension
        )

    def __getitem__(self, entity: str) -> Optional[str]:
        return self._entities[entity] if entity in self._entities else None

    def __setitem__(
        self, entity: str, value: str
    ) -> None:  # would puke with some exception if already known
        return self.set(entity, value, overwrite=False)

    def set(self, entity: str, value: str, overwrite: bool = True) -> None:
        if entity not in self._entities:
            # just set it; no complains here
            self._entities[entity] = value
        elif overwrite:
            lgr.warning(
                "Overwriting the entity %s from %s to %s for file %s",
                str(entity),
                str(self[entity]),
                str(value),
                self.__str__(),
            )
            self._entities[entity] = value
        else:
            # if it already exists, and overwrite is false:
            lgr.warning(
                "Setting the entity %s to %s for file %s failed",
                str(entity),
                str(value),
                self.__str__(),
            )

    @property  # as needed make them RW
    def suffix(self) -> str:
        return self._suffix

    @property
    def extension(self) -> Optional[str]:
        return self._extension


def sanitize_label(label: str) -> str:
    """Strips any non-BIDS compliant characters within label

    Parameters
    ----------
    label : string

    Returns
    -------
    clean_label : string
        New, sanitized label
    """
    clean_label = "".join(x for x in label if x.isalnum())
    if not clean_label:
        raise ValueError(
            "Label became empty after cleanup.  Please manually provide "
            "a suitable alphanumeric label."
        )
    if clean_label != label:
        lgr.warning(
            "%r label contained non-alphanumeric character(s), it "
            "was cleaned to be %r",
            label,
            clean_label,
        )
    return clean_label



#!/usr/bin/env python
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the Heudiconv package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##


def main():
    import os.path as op

    from setuptools import find_packages, findall, setup

    thispath = op.dirname(__file__)
    ldict = locals()

    # Get version and release info, which is all stored in heudiconv/info.py
    info_file = op.join(thispath, "heudiconv", "info.py")
    with open(info_file) as infofile:
        exec(infofile.read(), globals(), ldict)

    try:
        import versioningit  # noqa: F401
    except ImportError:
        # versioningit isn't installed; assume we're building a Debian package
        # from an sdist on an older Debian that doesn't support pybuild
        vglobals = {}
        with open(op.join(op.dirname(__file__), "heudiconv", "_version.py")) as fp:
            exec(fp.read(), vglobals)
        kwargs = {"version": vglobals["__version__"]}
    else:
        kwargs = {}

    def findsome(subdir, extensions):
        """Find files under subdir having specified extensions

        Leading directory (datalad) gets stripped
        """
        return [
            f.split(op.sep, 1)[1]
            for f in findall(subdir)
            if op.splitext(f)[-1].lstrip(".") in extensions
        ]

    # Only recentish versions of find_packages support include
    # heudiconv_pkgs = find_packages('.', include=['heudiconv*'])
    # so we will filter manually for maximal compatibility
    heudiconv_pkgs = [pkg for pkg in find_packages(".") if pkg.startswith("heudiconv")]

    setup(
        name=ldict["__packagename__"],
        author=ldict["__author__"],
        # author_email="team@???",
        description=ldict["__description__"],
        long_description=ldict["__longdesc__"],
        license=ldict["__license__"],
        classifiers=ldict["CLASSIFIERS"],
        packages=heudiconv_pkgs,
        entry_points={
            "console_scripts": [
                "heudiconv=heudiconv.cli.run:main",
                "heudiconv_monitor=heudiconv.cli.monitor:main",
            ]
        },
        python_requires=ldict["PYTHON_REQUIRES"],
        install_requires=ldict["REQUIRES"],
        extras_require=ldict["EXTRA_REQUIRES"],
        package_data={
            "heudiconv": ["py.typed"],
            "heudiconv.tests": [
                op.join("data", "*.dcm"),
                op.join("data", "*", "*.dcm"),
                op.join("data", "*", "*", "*.dcm"),
                op.join("data", "sample_nifti*"),
            ],
        },
        **kwargs,
    )


if __name__ == "__main__":
    main()



"""Heuristic to extract a b-value=0 DWI image (basically, a SE-EPI)
both as a fmap and as dwi

It is used just to test that a 'DIFFUSION' image that the user
chooses to extract as fmap (pepolar case) doesn't produce _bvecs/
_bvals json files, while it does for dwi images
"""

from __future__ import annotations

from typing import Optional

from heudiconv.utils import SeqInfo


def create_key(
    template: Optional[str],
    outtype: tuple[str, ...] = ("nii.gz",),
    annotation_classes: None = None,
) -> tuple[str, tuple[str, ...], None]:
    if template is None or not template:
        raise ValueError("Template must be a valid format string")
    return (template, outtype, annotation_classes)


def infotodict(
    seqinfo: list[SeqInfo],
) -> dict[tuple[str, tuple[str, ...], None], list[str]]:
    """Heuristic evaluator for determining which runs belong where

    allowed template fields - follow python string module:

    item: index within category
    subject: participant id
    seqitem: run number during scanning
    subindex: sub index within group
    """
    fmap = create_key("sub-{subject}/fmap/sub-{subject}_acq-b0dwi_epi")
    dwi = create_key("sub-{subject}/dwi/sub-{subject}_acq-b0dwi_dwi")

    info: dict[tuple[str, tuple[str, ...], None], list[str]] = {fmap: [], dwi: []}
    for s in seqinfo:
        if "DIFFUSION" in s.image_type:
            info[fmap].append(s.series_id)
            info[dwi].append(s.series_id)
    return info



from __future__ import annotations

import atexit
from collections import defaultdict
from collections.abc import ItemsView, Iterable, Iterator
from glob import glob
import logging
import os
import os.path as op
import re
import shutil
from types import ModuleType
from typing import Optional

from .dicoms import group_dicoms_into_seqinfos
from .utils import SeqInfo, StudySessionInfo, TempDirs, docstring_parameter

lgr = logging.getLogger(__name__)
tempdirs = TempDirs()
# Ensure they are cleaned up upon exit
atexit.register(tempdirs.cleanup)

_VCS_REGEX = r"%s\.(?:git|gitattributes|svn|bzr|hg)(?:%s|$)" % (op.sep, op.sep)

_UNPACK_FORMATS = tuple(sum((x[1] for x in shutil.get_unpack_formats()), []))


@docstring_parameter(_VCS_REGEX)
def find_files(
    regex: str,
    topdir: list[str] | tuple[str, ...] | str = op.curdir,
    exclude: Optional[str] = None,
    exclude_vcs: bool = True,
    dirs: bool = False,
) -> Iterator[str]:
    """Generator to find files matching regex

    Parameters
    ----------
    regex: string
    exclude: string, optional
      Matches to exclude
    exclude_vcs:
      If True, excludes commonly known VCS subdirectories.  If string, used
      as regex to exclude those files (regex: `{}`)
    topdir: string or list, optional
      Directory where to search
    dirs: bool, optional
      Either to match directories as well as files
    """
    if isinstance(topdir, (list, tuple)):
        for topdir_ in topdir:
            yield from find_files(
                regex,
                topdir=topdir_,
                exclude=exclude,
                exclude_vcs=exclude_vcs,
                dirs=dirs,
            )
        return
    for dirpath, dirnames, filenames in os.walk(topdir):
        names = (dirnames + filenames) if dirs else filenames
        paths = (op.join(dirpath, name) for name in names)
        for path in filter(re.compile(regex).search, paths):
            path = path.rstrip(op.sep)
            if exclude and re.search(exclude, path):
                continue
            if exclude_vcs and re.search(_VCS_REGEX, path):
                continue
            yield path


def get_extracted_dicoms(fl: Iterable[str]) -> ItemsView[Optional[str], list[str]]:
    """Given a collection of files and/or directories, list out and possibly
    extract the contents from archives.

    Parameters
    ----------
    fl
        Files (possibly archived) to process.

    Returns
    -------
    ItemsView[str | None, list[str]]
        The absolute paths of (possibly newly extracted) files.

    Notes
    -----
    For 'classical' heudiconv, if multiple archives are provided, they
    correspond to different sessions, so here we would group into sessions
    and return pairs `sessionid`, `files`  with `sessionid` being None if no
    "sessions" detected for that file or there was just a single tarball in the
    list.

    When contents of fl appear to be an unpackable archive, the contents are
    extracted into utils.TempDirs(f'heudiconvDCM') and the mode of all
    extracted files is set to 700.

    When contents of fl are a list of unarchived files, they are treated as
    a single session.

    When contents of fl are a list of unarchived and archived files, the
    unarchived files are grouped into a single session (key: None). If there is
    only one archived file, the contents of that file are grouped with
    the unarchived file. If there are multiple archived files, they are grouped
    into separate sessions.
    """
    sessions: dict[Optional[str], list[str]] = defaultdict(list)

    # keep track of session manually to ensure that the variable is bound
    # when it is used after the loop (e.g., consider situation with
    # fl being empty)
    session = 0

    # needs sorting to keep the generated "session" label deterministic
    for _, t in enumerate(sorted(fl)):
        if not t.endswith(_UNPACK_FORMATS):
            sessions[None].append(t)
            continue

        # Each file extracted must be associated with the proper session,
        # but the high-level shutil does not have a way to list the files
        # contained within each archive. So, files are temporarily
        # extracted into unique tempdirs
        # cannot use TempDirs since will trigger cleanup with __del__
        tmpdir = tempdirs(prefix="heudiconvDCM")

        # check content and sanitize permission bits before extraction
        os.chmod(tmpdir, mode=0o700)
        shutil.unpack_archive(t, extract_dir=tmpdir)

        archive_content = list(find_files(regex=".*", topdir=tmpdir))

        # may be too cautious (tmpdir is already 700).
        for f in archive_content:
            os.chmod(f, mode=0o700)
        # store full paths to each file, so we don't need to drag along
        # tmpdir as some basedir
        sessions[str(session)] = archive_content
        session += 1

    if session == 1:
        # we had only 1 session (and at least 1), so not really multiple
        # sessions according to classical 'heudiconv' assumptions, thus
        # just move them all into None
        sessions[None] += sessions.pop("0")

    return sessions.items()


def get_study_sessions(
    dicom_dir_template: Optional[str],
    files_opt: Optional[list[str]],
    heuristic: ModuleType,
    outdir: str,
    session: Optional[str],
    sids: Optional[list[str]],
    grouping: str = "studyUID",
) -> dict[StudySessionInfo, list[str] | dict[SeqInfo, list[str]]]:
    """Sort files or dicom seqinfos into study_sessions.

    study_sessions put together files for a single session of a subject
    in a study.  Two major possible workflows:

    - if dicom_dir_template provided -- doesn't pre-load DICOMs and just
      loads files pointed by each subject and possibly sessions as corresponding
      to different tarballs.

    - if files_opt is provided, sorts all DICOMs it can find under those paths
    """
    study_sessions: dict[StudySessionInfo, list[str] | dict[SeqInfo, list[str]]] = {}
    if dicom_dir_template:
        dicom_dir_template = op.abspath(dicom_dir_template)

        # MG - should be caught by earlier checks
        # assert not files_opt  # see above TODO
        assert sids
        # expand the input template

        if "{subject}" not in dicom_dir_template:
            raise ValueError(
                "dicom dir template must have {subject} as a placeholder for a "
                "subject id.  Got %r" % dicom_dir_template
            )
        for sid in sids:
            sdir = dicom_dir_template.format(subject=sid, session=session)
            for session_, files_ in get_extracted_dicoms(sorted(glob(sdir))):
                if session_ is not None and session:
                    lgr.warning(
                        "We had session specified (%s) but while analyzing "
                        "files got a new value %r (using it instead)"
                        % (session, session_)
                    )
                # in this setup we do not care about tracking "studies" so
                # locator would be the same None
                study_sessions[
                    StudySessionInfo(
                        None, session_ if session_ is not None else session, sid
                    )
                ] = files_
    else:
        # MG - should be caught on initial run
        # YOH - what if it is the initial run?
        # prep files
        assert files_opt
        files: list[str] = []
        for f in files_opt:
            if op.isdir(f):
                files += sorted(
                    find_files(".*", topdir=f, exclude_vcs=True, exclude=r"/\.datalad/")
                )
            else:
                files.append(f)

        # in this scenario we don't care about sessions obtained this way
        extracted_files: list[str] = []
        for _, files_ex in get_extracted_dicoms(files):
            extracted_files += files_ex

        # sort all DICOMS using heuristic
        seqinfo_dict = group_dicoms_into_seqinfos(
            extracted_files,
            grouping,
            file_filter=getattr(heuristic, "filter_files", None),
            dcmfilter=getattr(heuristic, "filter_dicom", None),
            custom_grouping=getattr(heuristic, "grouping", None),
            custom_seqinfo=getattr(heuristic, "custom_seqinfo", None),
        )

        if sids:
            if len(sids) != 1:
                raise RuntimeError(
                    "We were provided some subjects (%s) but "
                    "we can deal only "
                    "with overriding only 1 subject id. Got %d subjects and "
                    "found %d sequences" % (sids, len(sids), len(seqinfo_dict))
                )
            sid = sids[0]
        else:
            sid = None

        if not getattr(heuristic, "infotoids", None):
            # allow bypass with subject override
            if not sid:
                raise NotImplementedError(
                    "Cannot guarantee subject id - add "
                    "`infotoids` to heuristic file or "
                    "provide `--subjects` option"
                )
            lgr.warning(
                "Heuristic is missing an `infotoids` method, assigning "
                "empty method and using provided subject id %s. "
                "Provide `session` and `locator` fields for best results.",
                sid,
            )

            def infotoids(
                seqinfos: Iterable[SeqInfo], outdir: str  # noqa: U100
            ) -> dict[str, Optional[str]]:
                return {"locator": None, "session": None, "subject": None}

            heuristic.infotoids = infotoids  # type: ignore[attr-defined]

        for _studyUID, seqinfo in seqinfo_dict.items():
            # so we have a single study, we need to figure out its
            # locator, session, subject
            # TODO: Try except to ignore those we can't handle?
            # actually probably there should be a dedicated exception for
            # heuristics to throw if they detect that the study they are given
            # is not the one they would be willing to work on
            ids = heuristic.infotoids(seqinfo.keys(), outdir=outdir)
            # TODO:  probably infotoids is doomed to do more and possibly
            # split into multiple sessions!!!! but then it should be provided
            # full seqinfo with files which it would place into multiple groups
            study_session_info = StudySessionInfo(
                ids.get("locator"),
                ids.get("session", session) or session,
                sid or ids.get("subject", None),
            )
            lgr.info("Study session for %r", study_session_info)

            if grouping != "all":
                assert study_session_info not in study_sessions, (
                    f"Existing study session {study_session_info} "
                    f"already in analyzed sessions {study_sessions.keys()}"
                )
            study_sessions[study_session_info] = seqinfo
    return study_sessions



from __future__ import annotations

from typing import Optional

from heudiconv.utils import SeqInfo


def create_key(
    template: Optional[str],
    outtype: tuple[str, ...] = ("nii.gz",),
    annotation_classes: None = None,
) -> tuple[str, tuple[str, ...], None]:
    if template is None or not template:
        raise ValueError("Template must be a valid format string")
    return (template, outtype, annotation_classes)


def infotodict(
    seqinfo: list[SeqInfo],
) -> dict[tuple[str, tuple[str, ...], None], list]:
    """Heuristic evaluator for determining which runs belong where

    allowed template fields - follow python string module:

    item: index within category
    subject: participant id
    seqitem: run number during scanning
    subindex: sub index within group
    """
    t1w = create_key("anat/sub-{subject}_T1w")
    t2w = create_key("anat/sub-{subject}_acq-{acq}_T2w")
    flair = create_key("anat/sub-{subject}_acq-{acq}_FLAIR")
    rest = create_key("func/sub-{subject}_task-rest_acq-{acq}_run-{item:02d}_bold")

    info: dict[tuple[str, tuple[str, ...], None], list] = {
        t1w: [],
        t2w: [],
        flair: [],
        rest: [],
    }

    for seq in seqinfo:
        x, _, z, n_vol, protocol, dcm_dir = (
            seq.dim1,
            seq.dim2,
            seq.dim3,
            seq.dim4,
            seq.protocol_name,
            seq.dcm_dir_name,
        )
        # t1_mprage --> T1w
        if (
            (z == 160)
            and (n_vol == 1)
            and ("t1_mprage" in protocol)
            and ("XX" not in dcm_dir)
        ):
            info[t1w] = [seq.series_id]
        # t2_tse --> T2w
        if (
            (z == 35)
            and (n_vol == 1)
            and ("t2_tse" in protocol)
            and ("XX" not in dcm_dir)
        ):
            info[t2w].append({"item": seq.series_id, "acq": "TSE"})
        # T2W --> T2w
        if (
            (z == 192)
            and (n_vol == 1)
            and ("T2W" in protocol)
            and ("XX" not in dcm_dir)
        ):
            info[t2w].append({"item": seq.series_id, "acq": "highres"})
        # t2_tirm --> FLAIR
        if (
            (z == 35)
            and (n_vol == 1)
            and ("t2_tirm" in protocol)
            and ("XX" not in dcm_dir)
        ):
            info[flair].append({"item": seq.series_id, "acq": "TIRM"})
        # t2_flair --> FLAIR
        if (
            (z == 160)
            and (n_vol == 1)
            and ("t2_flair" in protocol)
            and ("XX" not in dcm_dir)
        ):
            info[flair].append({"item": seq.series_id, "acq": "highres"})
        # T2FLAIR --> FLAIR
        if (
            (z == 192)
            and (n_vol == 1)
            and ("T2-FLAIR" in protocol)
            and ("XX" not in dcm_dir)
        ):
            info[flair].append({"item": seq.series_id, "acq": "highres"})
        # EPI (physio-matched) --> bold
        if (
            (x == 128)
            and (z == 28)
            and (n_vol == 300)
            and ("EPI" in protocol)
            and ("XX" not in dcm_dir)
        ):
            info[rest].append({"item": seq.series_id, "acq": "128px"})
        # EPI (physio-matched_NEW) --> bold
        if (
            (x == 64)
            and (z == 34)
            and (n_vol == 300)
            and ("EPI" in protocol)
            and ("XX" not in dcm_dir)
        ):
            info[rest].append({"item": seq.series_id, "acq": "64px"})
    return info



"""Testing conversion with conversion saved on datalad"""
from __future__ import annotations

from glob import glob
import os
import os.path as op
from pathlib import Path
import re
from typing import Optional

import pydicom as dcm
import pytest

from heudiconv.cli.run import main as runner
from heudiconv.parser import find_files
from heudiconv.utils import load_json

# testing utilities
from .utils import TESTS_DATA_PATH, fetch_data, gen_heudiconv_args

have_datalad = True
try:
    from datalad.support.exceptions import IncompleteResultsError
except ImportError:
    have_datalad = False


@pytest.mark.skipif(not have_datalad, reason="no datalad")
@pytest.mark.parametrize("subject", ["sid000143"])
@pytest.mark.parametrize("heuristic", ["reproin.py"])
@pytest.mark.parametrize("anon_cmd", [None, "anonymize_script.py"])
def test_conversion(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
    subject: str,
    heuristic: str,
    anon_cmd: Optional[str],
) -> None:
    monkeypatch.chdir(tmp_path)
    try:
        datadir = fetch_data(
            tmp_path,
            "dbic/QA",  # path from datalad database root
            getpath=op.join("sourcedata", f"sub-{subject}"),
        )
    except IncompleteResultsError as exc:
        pytest.skip("Failed to fetch test data: %s" % str(exc))
    outdir = tmp_path / "out"
    outdir.mkdir()

    args = gen_heudiconv_args(
        datadir,
        str(outdir),
        subject,
        heuristic,
        anon_cmd,
        template="sourcedata/sub-{subject}/*/*/*.tgz",
    )
    runner(args)  # run conversion

    # Get the possibly anonymized subject id and verify that it was
    # anonymized or not:
    subjects_maybe_anon = glob(f"{outdir}/sub-*")
    assert len(subjects_maybe_anon) == 1  # just one should be there
    subject_maybe_anon = op.basename(subjects_maybe_anon[0])[4:]

    if anon_cmd:
        assert subject_maybe_anon != subject
    else:
        assert subject_maybe_anon == subject

    # verify functionals were converted
    outfiles = sorted(
        [
            f[len(str(outdir)) :]
            for f in glob(f"{outdir}/sub-{subject_maybe_anon}/func/*")
        ]
    )
    assert outfiles
    datafiles = sorted(
        [f[len(datadir) :] for f in glob(f"{datadir}/sub-{subject}/ses-*/func/*")]
    )
    # original data has ses- but because we are converting only func, and not
    # providing any session, we will not "match". Let's strip away the session
    datafiles = [re.sub(r"[/\\_]ses-[^/\\_]*", "", f) for f in datafiles]
    if not anon_cmd:
        assert outfiles == datafiles
    else:
        assert outfiles != datafiles  # sid was anonymized
        assert len(outfiles) == len(datafiles)  # but we have the same number of files

    # compare some json metadata
    json_ = "{}/task-rest_acq-24mm64sl1000tr32te600dyn_bold.json".format
    orig, conv = (load_json(json_(datadir)), load_json(json_(outdir)))
    keys = ["EchoTime", "MagneticFieldStrength", "Manufacturer", "SliceTiming"]
    for key in keys:
        assert orig[key] == conv[key]


@pytest.mark.skipif(not have_datalad, reason="no datalad")
def test_multiecho(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
    subject: str = "MEEPI",
    heuristic: str = "bids_ME.py",
) -> None:
    monkeypatch.chdir(tmp_path)
    try:
        datadir = fetch_data(tmp_path, "dicoms/velasco/MEEPI")
    except IncompleteResultsError as exc:
        pytest.skip("Failed to fetch test data: %s" % str(exc))

    outdir = tmp_path / "out"
    outdir.mkdir()
    args = gen_heudiconv_args(datadir, str(outdir), subject, heuristic)
    runner(args)  # run conversion

    # check if we have echo functionals
    echoes = glob(op.join("out", "sub-" + subject, "func", "*echo*nii.gz"))
    assert len(echoes) == 3

    # check EchoTime of each functional
    # ET1 < ET2 < ET3
    prev_echo = 0
    for echo in sorted(echoes):
        _json = echo.replace(".nii.gz", ".json")
        assert _json
        echotime = load_json(_json).get("EchoTime", None)
        assert echotime > prev_echo
        prev_echo = echotime

    events = glob(op.join("out", "sub-" + subject, "func", "*events.tsv"))
    for event in events:
        assert "echo-" not in event


@pytest.mark.parametrize("subject", ["merged"])
def test_grouping(tmp_path: Path, subject: str) -> None:
    dicoms = [op.join(TESTS_DATA_PATH, fl) for fl in ["axasc35.dcm", "phantom.dcm"]]
    # ensure DICOMs are different studies
    studyuids = {
        dcm.dcmread(fl, stop_before_pixels=True).StudyInstanceUID for fl in dicoms
    }
    assert len(studyuids) == len(dicoms)
    # symlink to common location
    outdir = tmp_path / "out"
    outdir.mkdir()
    datadir = tmp_path / subject
    datadir.mkdir()
    for fl in dicoms:
        os.symlink(fl, datadir / op.basename(fl))

    template = op.join("{subject}/*.dcm")
    hargs = gen_heudiconv_args(
        str(tmp_path), str(outdir), subject, "convertall.py", template=template
    )

    with pytest.raises(AssertionError):
        runner(hargs)

    # group all found DICOMs under subject, despite conflicts
    hargs += ["-g", "all"]
    runner(hargs)
    assert (
        len(
            list(
                find_files(
                    rf"(^|{re.escape(os.sep)})run0",
                    str(outdir),
                    exclude_vcs=False,
                    dirs=True,
                )
            )
        )
        == 4
    )
    tsv = outdir / "participants.tsv"
    assert tsv.exists()
    lines = tsv.open().readlines()
    assert len(lines) == 2
    assert lines[1].split("\t")[0] == "sub-{}".format(subject)



"""A demo convertall heuristic with custom_seqinfo extracting affine and sample DICOM path

This heuristic also demonstrates on how to create a "derived" heuristic which would augment
behavior of an already existing heuristic without complete rewrite.  Such approach could be
useful for heuristic like  reproin  to overload mapping etc.
"""
from __future__ import annotations

from typing import Any

import nibabel.nicom.dicomwrappers as dw

from .convertall import *  # noqa: F403


def custom_seqinfo(
    series_files: list[str], wrapper: dw.Wrapper, **kw: Any  # noqa: U100
) -> tuple[str | None, str]:
    """Demo for extracting custom header fields into custom_seqinfo field

    Operates on already loaded DICOM data.
    Origin: https://github.com/nipy/heudiconv/pull/333
    """

    from nibabel.nicom.dicomwrappers import WrapperError

    try:
        affine = str(wrapper.affine)
    except WrapperError:
        lgr.exception("Errored out while obtaining/converting affine")  # noqa: F405
        affine = None
    return affine, series_files[0]



from __future__ import annotations

from typing import Optional

from heudiconv.utils import SeqInfo


def create_key(
    template: Optional[str],
    outtype: tuple[str, ...] = ("nii.gz", "dicom"),
    annotation_classes: None = None,
) -> tuple[str, tuple[str, ...], None]:
    if template is None or not template:
        raise ValueError("Template must be a valid format string")
    return (template, outtype, annotation_classes)


def infotodict(
    seqinfo: list[SeqInfo],
) -> dict[tuple[str, tuple[str, ...], None], list]:
    """Heuristic evaluator for determining which runs belong where

    allowed template fields - follow python string module:

    item: index within category
    subject: participant id
    seqitem: run number during scanning
    subindex: sub index within group
    """
    t1 = create_key("anat/sub-{subject}_T1w")
    t2 = create_key("anat/sub-{subject}_T2w")
    rest = create_key("func/sub-{subject}_dir-{acq}_task-rest_run-{item:02d}_bold")
    face = create_key("func/sub-{subject}_task-face_run-{item:02d}_acq-{acq}_bold")
    gamble = create_key(
        "func/sub-{subject}_task-gambling_run-{item:02d}_acq-{acq}_bold"
    )
    conflict = create_key(
        "func/sub-{subject}_task-conflict_run-{item:02d}_acq-{acq}_bold"
    )
    dwi = create_key("dwi/sub-{subject}_dir-{acq}_run-{item:02d}_dwi")

    fmap_rest = create_key(
        "fmap/sub-{subject}_acq-func{acq}_dir-{dir}_run-{item:02d}_epi"
    )
    fmap_dwi = create_key(
        "fmap/sub-{subject}_acq-dwi{acq}_dir-{dir}_run-{item:02d}_epi"
    )

    info: dict[tuple[str, tuple[str, ...], None], list] = {
        t1: [],
        t2: [],
        rest: [],
        face: [],
        gamble: [],
        conflict: [],
        dwi: [],
        fmap_rest: [],
        fmap_dwi: [],
    }

    for idx, s in enumerate(seqinfo):
        if (s.dim3 == 208) and (s.dim4 == 1) and ("T1w" in s.protocol_name):
            info[t1] = [s.series_id]
        if (s.dim3 == 208) and ("T2w" in s.protocol_name):
            info[t2] = [s.series_id]
        if (s.dim4 >= 99) and (
            ("dMRI_dir98_AP" in s.protocol_name) or ("dMRI_dir99_AP" in s.protocol_name)
        ):
            acq = s.protocol_name.split("dMRI_")[1].split("_")[0] + "AP"
            info[dwi].append({"item": s.series_id, "acq": acq})
        if (s.dim4 >= 99) and (
            ("dMRI_dir98_PA" in s.protocol_name) or ("dMRI_dir99_PA" in s.protocol_name)
        ):
            acq = s.protocol_name.split("dMRI_")[1].split("_")[0] + "PA"
            info[dwi].append({"item": s.series_id, "acq": acq})
        if (s.dim4 == 1) and (
            ("dMRI_dir98_AP" in s.protocol_name) or ("dMRI_dir99_AP" in s.protocol_name)
        ):
            acq = s.protocol_name.split("dMRI_")[1].split("_")[0]
            info[fmap_dwi].append({"item": s.series_id, "dir": "AP", "acq": acq})
        if (s.dim4 == 1) and (
            ("dMRI_dir98_PA" in s.protocol_name) or ("dMRI_dir99_PA" in s.protocol_name)
        ):
            acq = s.protocol_name.split("dMRI_")[1].split("_")[0]
            info[fmap_dwi].append({"item": s.series_id, "dir": "PA", "acq": acq})
        if (s.dim4 == 420) and ("rfMRI_REST_AP" in s.protocol_name):
            info[rest].append({"item": s.series_id, "acq": "AP"})
        if (s.dim4 == 420) and ("rfMRI_REST_PA" in s.protocol_name):
            info[rest].append({"item": s.series_id, "acq": "PA"})
        if (s.dim4 == 1) and ("rfMRI_REST_AP" in s.protocol_name):
            if seqinfo[idx + 1][9] != 420:
                continue
            info[fmap_rest].append({"item": s.series_id, "dir": "AP", "acq": ""})
        if (s.dim4 == 1) and ("rfMRI_REST_PA" in s.protocol_name):
            info[fmap_rest].append({"item": s.series_id, "dir": "PA", "acq": ""})
        if (s.dim4 == 346) and ("tfMRI_faceMatching_AP" in s.protocol_name):
            info[face].append({"item": s.series_id, "acq": "AP"})
        if (s.dim4 == 346) and ("tfMRI_faceMatching_PA" in s.protocol_name):
            info[face].append({"item": s.series_id, "acq": "PA"})
        if (s.dim4 == 288) and ("tfMRI_conflict_AP" in s.protocol_name):
            info[conflict].append({"item": s.series_id, "acq": "AP"})
        if (s.dim4 == 288) and ("tfMRI_conflict_PA" in s.protocol_name):
            info[conflict].append({"item": s.series_id, "acq": "PA"})
        if (s.dim4 == 223) and ("tfMRI_gambling_AP" in (s.protocol_name)):
            info[gamble].append({"item": s.series_id, "acq": "AP"})
        if (s.dim4 == 223) and ("tfMRI_gambling_PA" in s.protocol_name):
            info[gamble].append({"item": s.series_id, "acq": "PA"})
    return info



from __future__ import annotations

import logging
from typing import Optional

from heudiconv.utils import SeqInfo

lgr = logging.getLogger("heudiconv")


def create_key(
    template: Optional[str],
    outtype: tuple[str, ...] = ("nii.gz",),
    annotation_classes: None = None,
) -> tuple[str, tuple[str, ...], None]:
    if template is None or not template:
        raise ValueError("Template must be a valid format string")
    return (template, outtype, annotation_classes)


def infotodict(
    seqinfo: list[SeqInfo],
) -> dict[tuple[str, tuple[str, ...], None], list[str]]:
    """Heuristic evaluator for determining which runs belong where

    allowed template fields - follow python string module:

    item: index within category
    subject: participant id
    seqitem: run number during scanning
    subindex: sub index within group
    """

    data = create_key("run{item:03d}")
    info: dict[tuple[str, tuple[str, ...], None], list[str]] = {data: []}

    for s in seqinfo:
        """
        The namedtuple `s` contains the following fields:

        * total_files_till_now
        * example_dcm_file
        * series_id
        * dcm_dir_name
        * unspecified2
        * unspecified3
        * dim1
        * dim2
        * dim3
        * dim4
        * TR
        * TE
        * protocol_name
        * is_motion_corrected
        * is_derived
        * patient_id
        * study_description
        * referring_physician_name
        * series_description
        * image_type
        """

        info[data].append(s.series_id)
    return info



from __future__ import annotations

from typing import Optional

from heudiconv.utils import SeqInfo

scaninfo_suffix = ".json"


def create_key(
    template: Optional[str],
    outtype: tuple[str, ...] = ("nii.gz",),
    annotation_classes: None = None,
) -> tuple[str, tuple[str, ...], None]:
    if template is None or not template:
        raise ValueError("Template must be a valid format string")
    return (template, outtype, annotation_classes)


def infotodict(
    seqinfo: list[SeqInfo],
) -> dict[tuple[str, tuple[str, ...], None], list[str]]:
    """Heuristic evaluator for determining which runs belong where

    allowed template fields - follow python string module:

    item: index within category
    subject: participant id
    seqitem: run number during scanning
    subindex: sub index within group
    """

    label_map = {
        "movie": "movielocalizer",
        "retmap": "retmap",
        "visloc": "objectcategories",
    }
    info: dict[tuple[str, tuple[str, ...], None], list[str]] = {}
    for s in seqinfo:
        if "EPI_3mm" not in s.protocol_name:
            continue
        label = s.protocol_name.split("_")[2].split()[0].strip("1234567890").lower()
        if label in ("movie", "retmap", "visloc"):
            key = create_key(
                "ses-localizer/func/{subject}_ses-localizer_task-%s_run-{item:01d}_bold"
                % label_map[label]
            )
        elif label == "sense":
            # pilot retmap had different description
            key = create_key(
                "ses-localizer/func/{subject}_ses-localizer_task-retmap_run-{item:01d}_bold"
            )
        elif label == "r":
            key = create_key(
                "ses-movie/func/{subject}_ses-movie_task-movie_run-%i_bold"
                % int(s.protocol_name.split("_")[2].split()[0][-1])
            )
        else:
            raise RuntimeError("YOU SHALL NOT PASS!")

        if key not in info:
            info[key] = []

        info[key].append(s.series_id)

    return info
